/////////////////////////////////////////////////////////////
//                                                         //
// graph_cutpoint.cpp -- this file contains implementation //
//           of piecewise quadratic optimization           // 
// ver 1.01                                                //  
//                                                         //   
// Language: standard C++                                  // 
// Platform: Linux Debian                                  //
// Contact:  Minmei Hou, PhD Candidate                     //  
//           CSE Dept., Penn State University              //
//           mhou@cse.psu.edu                              //    
/////////////////////////////////////////////////////////////


#include <iostream>
#include <fstream>
#include <vector>
#include <algorithm>
#include "bzTag.hpp"
#include "graph_cutpoint.hpp"

using namespace std;


bool compar_line_start(line* firstline, line* secondline) {
  return firstline->start < secondline->start;
}

bool compar_point(point* p1, point* p2) {
  return p1->pos < p2->pos;
}

posShape* getMinimumPoint(vector<long>& computePositions, long left, long right) {
    long sum=0, ave;
    posShape* ptr;
    ptr = new posShape;

    int i, pointCount=computePositions.size();

    if ( computePositions.size() == 0 ) {
      ptr->pos = left/2 + right/2;
      ptr->shape = 'd';
      return ptr;
    }

    for (i=0; i<pointCount; i++)
      sum += computePositions[i];
    ave = sum/pointCount;


    if ( ave < left) {
      ptr->pos = left;
      ptr->shape = 'i';
    }
    else if ( ave > right ) {
      ptr->pos = right;
      ptr->shape = 'd';
    }
    else {
      ptr->pos = ave;
      ptr->shape = 'b';
    }
    return ptr;
}

posShape* processGapCase(vector<long>& starts, vector<long>& ends, long start_range, long end_range, vector<long>& gaps_starts, vector<long>& gaps_ends) {
  int i, gapSize=gaps_ends.size(), pointCount=starts.size();
  int gap_len=0, max_gap_index=-1;
  long inner_range_start, inner_range_end;
  vector<long> change_points;
  posShape* aPosShape;

  // find the largest gap 
  for (i=0; i<gapSize; i++)
    if ( (gaps_ends[i]-gaps_starts[i]+1) > gap_len ) {
      gap_len = gaps_ends[i]-gaps_starts[i]+1;
      max_gap_index = i;
    }
  //cerr << "gap index: " << max_gap_index << " gap length: " << gap_len << endl;
  inner_range_start = gaps_starts[max_gap_index];
  inner_range_end   = gaps_ends[max_gap_index];

  //cerr << "inner ranges: " << inner_range_start << "  " << inner_range_end << endl;
  for (i=0; i<pointCount; i++)
    if ( starts[i] > inner_range_end && starts[i] < end_range )
      change_points.push_back(starts[i]);
  for (i=0; i<pointCount; i++)
    if ( ends[i] > start_range && ends[i] < inner_range_start )
      change_points.push_back(ends[i]);
  
  aPosShape = getMinimumPoint(change_points, inner_range_start, inner_range_end);
  //cerr << "cut point: " << aPosShape->pos << endl;
  return aPosShape;
}

void free_points(vector<point*>& points) {
  int size=points.size(), i;
  for (i=0; i<size; i++)
    delete points[i];
  points.clear();
}

void free_lines(vector<line*>& lines) {
  int size=lines.size(), i;
  for (i=0; i<size; i++)
    delete lines[i];
  lines.clear();
}

posShape* processNoGapCase(vector<long>& starts, vector<long>& ends, long start_range, long end_range) {
    vector<char> shapes;
    vector<long> minimums;
    vector<line*> lines;
    vector<point*> points;
    vector<long> computePositions;
    point* newpoint;
    line* newline;
    posShape* aPosShape=NULL;
    int lineCount=starts.size(), i, k, posCount;
    long interval_left, interval_right;

    for (i=0; i<lineCount; i++) {
      newline = new line;
      newline->start = starts[i];
      newline->end   = ends[i];
      lines.push_back(newline);
      if ( starts[i] >= start_range && starts[i] <= end_range ) {
	newpoint = new point;
	newpoint->pos = starts[i];
	newpoint->property = 's';
	points.push_back(newpoint);
      }
      if ( ends[i] >= start_range && ends[i] <= end_range ) {
	newpoint = new point;
	newpoint->pos = ends[i];
	newpoint->property = 'e';
	points.push_back(newpoint);
      }
    }
    sort(points.begin(), points.end(), compar_point);
    sort(lines.begin(), lines.end(), compar_line_start); // replaced by interval tree

    // initialize computePositions
    for (i=0; i<lineCount; i++) 
      if ( lines[i]->start < start_range && lines[i]->end > start_range )
	computePositions.push_back(lines[i]->end);
    free_lines(lines);

    int pointCount = points.size();
    i=0;
    while ( i<pointCount && points[i]->pos <= start_range )
      i++;

    interval_left = start_range;
    for ( ; i<pointCount; ) {
      interval_right = points[i]->pos;
      aPosShape = getMinimumPoint(computePositions, interval_left, interval_right);
      if ( minimums.size() == 0 ) {
	minimums.push_back(aPosShape->pos);
	shapes.push_back(aPosShape->shape);
      }
      else {
	if ( *(shapes.end()-1) == 'd' ) { // prev decrease
	  if ( aPosShape->shape == 'd' ) { // decrease
	    *(minimums.end()-1) = aPosShape->pos;
	  }
	  else if ( aPosShape->shape == 'i' ) { // increase
	    *(shapes.end()-1) = 'b';
	  }
	  else { // bottom
	    *(minimums.end()-1) = aPosShape->pos;
	    *(shapes.end()-1) = 'b';
	  }
	}
	else if ( *(shapes.end()-1) == 'b' ) { // prev bottom
	  if ( aPosShape->shape == 'd' || aPosShape->shape == 'b') { // decrease or bottom
	    minimums.push_back(aPosShape->pos);
	    shapes.push_back(aPosShape->shape);
	  }
	}
	else { // prev increase
	  if ( aPosShape->shape == 'd' || aPosShape->shape == 'b') { // decrease
	    minimums.push_back(aPosShape->pos);
	    shapes.push_back(aPosShape->shape);
	  }
	}
      }

      while ( i<pointCount && points[i++]->pos == interval_right ) {
	if ( points[i-1]->property == 's' )
	  computePositions.push_back(interval_right);
	else {
	  posCount=computePositions.size();
	  for (k=posCount-1; k>=0; k--)
	    if ( computePositions[k] == interval_right ) {
	      computePositions.erase(computePositions.begin() + k);	
	      break;
	    }
	}
      }
      interval_left = interval_right;
    }
    if ( minimums.size() == 0 ){
      minimums.push_back( start_range/2 + end_range/2 );
      shapes.push_back('d');
      aPosShape = new posShape;
    }
    //cerr << "number of minimum points: " << minimums.size() << " first value: " << minimums[0] << endl;

    aPosShape->pos = minimums[0];
    aPosShape->shape = shapes[0];
    free_points(points);
    return aPosShape;
}

void getGaps(vector<long>& starts, vector<long>& ends, long &start_range, long &end_range, vector<long>& gaps_starts, vector<long>& gaps_ends) {
  long count, i, j;
  long min_start, max_end, length;
  char* markArr;

  count = starts.size();
  min_start = MAX_INT;
  for (i=0; i<count; i++)
    if ( min_start > starts[i] )
      min_start = starts[i];
  count = ends.size();
  max_end = 0;
  for (i=0; i<count; i++)
    if ( max_end < ends[i] )
      max_end = ends[i];

  if ( min_start > end_range )
    min_start = start_range;
  if ( max_end < start_range )
    max_end = end_range;

  length = max_end-min_start+1;

  if ( length <= 0 )
    return;
  
  markArr = (char*)malloc(length*sizeof(char));
  memset(markArr, 'n', length);

  /*
  for (i=0; i<count; i++)
    for (j=starts[i]-min_start; j<ends[i]-min_start+1; j++)
      markArr[j] = 'y';
  */

  if ( start_range < min_start )
    start_range = min_start;
  if ( end_range > max_end )
    end_range = max_end;

  int inside_gap = 0, gap_count=0;
  for (j=start_range-min_start; j<=end_range-min_start; j++) {
    if ( markArr[j] == 'n' ) {
      if ( inside_gap == 0 ) {
	inside_gap = 1;
	gap_count++;
	gaps_starts.push_back(j+min_start);
      }
    }
    else if ( inside_gap == 1 ) {
      inside_gap = 0;
      gaps_ends.push_back(j-1+min_start);
    }
  }
  free(markArr);
}

long graph_cutpoint(vector<long>& starts, vector<long>& ends, long start_range, long end_range) {
  vector<long> gaps_starts, gaps_ends;
  posShape* aPosShape;
  long retLong;
  
  getGaps(starts, ends, start_range, end_range, gaps_starts, gaps_ends);;
  
  //cerr << gaps_ends.size() << endl;
  
  if ( gaps_ends.size() > 0 ) {
    aPosShape = processGapCase( starts, ends, start_range, end_range, gaps_starts, gaps_ends);
  }
  else { // no gap at all
    aPosShape = processNoGapCase(starts, ends, start_range, end_range);
  }
  retLong = aPosShape->pos;
  delete aPosShape;
  return retLong;
}


//#define CUTPOINT_TEST
#ifdef CUTPOINT_TEST
int main(int argc, char** argv) {
  vector<long> starts, ends;
  long start, end, start_range, end_range;
  ifstream ifs;

  if ( argc < 4 ) {
    cerr << "args: point-file start-range end-range\n";
    return 1;
  }
  start_range = atol(argv[2]);
  end_range = atol(argv[3]);

  // reading in points
  ifs.open(argv[1]);
  ifs >> start;
  while ( ifs ) {
    starts.push_back(start);
    ifs >> end;
    ends.push_back(end);
    ifs >> start;
  }
  ifs.close();

  cerr << graph_cutpoint(starts, ends, start_range, end_range) << endl;

  return 0;
}

#endif
