///////////////////////////////////////////////////////
//                                                   //
// homologMiner.cpp -- a tool to look for duplicates //
//           in a genome and construct homologous    //
//           groups                                  //
// ver 1.01                                          //  
//                                                   //   
// Language: standard C++                            // 
// Platform: Linux Debian                            //
// Contact:  Minmei Hou, PhD Candidate               //  
//           CSE Dept., Penn State University        //
//           mhou@cse.psu.edu                        //    
///////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////
//  Build Process                                            //
///////////////////////////////////////////////////////////////
//  Required files:                                          //
//    bzTag.hpp bzTag.cpp graph_cutpoint.hpp graph_cutpoint.cpp 
//  graph_node.hpp graph_flow_min_cut_ll.hpp graph_flow_min_cut_ll.cpp 
//  homologMiner.cpp splitDupRepeat.hpp splitDupRepeat.cpp 
//  parameters.cpp                                           //
//                                                           //
//  compiler command:                                        //
//    g++ -g -Wall parameters.cpp bzTag.cpp graph_cutpoint.cpp 
//  graph_flow_min_cut_ll.cpp homologMiner.cpp splitDupRepeat.cpp 
//  -o HomologMiner                                          //
///////////////////////////////////////////////////////////////


/*
  Maintenance History:
  ====================
  ver 1.01: 07 Jan 2007
  - moved parameters to a single source file
  ver 1.00: 17 Nov 2006
  - first release
*/


#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <deque>
#include <list>
#include <iterator>
#include <string>
#include <algorithm>
#include "bzTag.hpp"
#include "graph_cutpoint.hpp"
#include "graph_node.hpp"
#include "graph_flow_min_cut_ll.hpp"
#include "splitDupRepeat.hpp"
using namespace std;

extern long* fakestarts;
extern int* init_chrs;

extern long CLOSE_THRESHOLD;
extern long CUT_THRESHOLD;
extern long SML_MIN_LENGTH;
extern long GAP_THRESHOLD;
extern long NODE_THRESHOLD;
extern long CHAIN_IGNORE_LENGTH;
extern int DUP_GAP;
extern long MAX_SPLIT_OVERLAP;
extern long CHAIN_SMALL_GAP;

extern long MIN_LENGTH;
extern long MIN_CHAIN;
extern int REFINE_ITERATION;
extern int SPLITMERGE_ITERATION;

long origY1, origY2, origX1;

//--------------------------------------------------//
//----------------- sorting criteria ---------------//
// sort from small to large
bool compar_tag_x1 (bzTag* tag1, bzTag* tag2) {
  return tag1->b1 < tag2->b1;
}

bool compar_tag_x2 (bzTag* tag1, bzTag* tag2) {
  return tag1->b2 < tag2->b2;
}

// sort from small to large
bool compar_node_start(node* n1, node* n2) {
  return n1->start < n2->start;
}

// since tag1 tag2 must locates within a same node, and within a certain 
// threshold, there is no overflow problem
// (origY1, origY2) locates left below
bool compar_tags_distancePos(bzTag* tag1, bzTag* tag2) {
  return ( ((tag1->b1-origY1)*(tag1->b1-origY1)+(tag1->b2-origY2)*(tag1->b2-origY2))
	   < ((tag2->b1-origY1)*(tag2->b1-origY1)+(tag2->b2-origY2)*(tag2->b2-origY2)) );
}

// (origX1, origX2) locates right below
bool compar_tags_distanceNeg(bzTag* tag1, bzTag* tag2) {
  return ( ((tag1->e1-origX1)*(tag1->e1-origX1)+(tag1->b2-origY2)*(tag1->b2-origY2))
	   < ((tag2->e1-origX1)*(tag2->e1-origX1)+(tag2->b2-origY2)*(tag2->b2-origY2)) );
}

bool compar_chain_x2( vector<bzTag*>* chain1, vector<bzTag*>* chain2) {
  return ( (*chain1)[0]->b2 < (*chain2)[0]->b2 );
}

bool compar_edge_x1(edge* e1, edge* e2) {
  bzTag* t1, *t2;
  t1 = *(e1->tags->begin());
  t2 = *(e2->tags->begin());
  return t1->b1 < t2->b1;
}

bool compar_edge_x2(edge* e1, edge* e2) {
  bzTag* t1, *t2;
  t1 = *(e1->tags->begin());
  t2 = *(e2->tags->begin());  
  return t1->b2 < t2->b2;
}

bool compar_int(int i1, int i2) {
  return i1 < i2;
}

bool compar_int_vec(vector<int>* vec1, vector<int>* vec2) {
  return  (*vec1)[0] < (*vec2)[0];
}
//-------------- end of sorting criteria -----------//
//--------------------------------------------------//

//-------- extract nodes from tags ---------------//
// assume tags are sorted according to x1         //
void getNodesFromTags(vector<bzTag*>* tags, vector<node*>& nodes) {
  int tagSize=tags->size(), i;
  long j, min_pos = MAX_INT, max_pos = 0, inputlength;
  bzTag* currTag;
  char* dupMarkArr;

  // determine the min and max points of the tags
  for (i=0; i<tagSize; i++) {
    currTag = (*tags)[i];
    if ( currTag->b1 < min_pos )
      min_pos = currTag->b1;
    if ( currTag->b2 < min_pos )
      min_pos = currTag->b2;
    if ( currTag->e1 > max_pos )
      max_pos = currTag->e1;
    if ( currTag->e2 > max_pos )
      max_pos = currTag->e2;
  }

  inputlength = max_pos - min_pos + 1;
  // mark the duplicated regions
  dupMarkArr = (char*)malloc( inputlength*sizeof(char));
  memset(dupMarkArr, 'n', inputlength);
  for (i=0; i<tagSize; i++) {
    currTag = (*tags)[i];
    memset(dupMarkArr+(currTag->b1-min_pos), 'd', currTag->e1-currTag->b1+1);
    memset(dupMarkArr+(currTag->b2-min_pos), 'd', currTag->e2-currTag->b2+1);
  }
  
  // construct nodes
  node* newnode;
  long last_node_start = -1, last_tag_end=MAX_INT;
  short insideTag=-1;

  //long last_beg;
  //long last_pos = -1;
  for (j = 0; j < inputlength; j++) {
    if ( insideTag == -1 && dupMarkArr[j] == 'd') {
      insideTag = 1;
      last_node_start = j;
    }
    else if ( insideTag == 1 && dupMarkArr[j] == 'n' ) {
      insideTag = 0;
      last_tag_end = j-1;
    }
    else if ( (insideTag == 0 && dupMarkArr[j] == 'd') || (j==max_pos-min_pos) ) {
      if ( dupMarkArr[j] == 'd' )
	insideTag = 1;

      if ( (j - last_tag_end > DUP_GAP) || (j==max_pos-min_pos && last_node_start > 0) ) {
	newnode = new node;
	newnode->start = last_node_start + min_pos;

	if ( j == inputlength - 1 && dupMarkArr[j] == 'd' )
	  newnode->end = j + min_pos;
	else
	  newnode->end = last_tag_end + min_pos;

	newnode->tags = new list<bzTag*>;
	newnode->edges = new list<edge*>;
	newnode->chr = 0;
	nodes.push_back(newnode);
	last_node_start = j;
      }// if 
    }//else if
  }// for
  free(dupMarkArr);
  sort(nodes.begin(), nodes.end(), compar_node_start);
}

//------------------------------------------------//
//------- displaying procedures ------------------//
// determine the max num of overlapped region
int determNumCopy(vector<eventDot*>& dots) {
  int max, copy, i, count = dots.size();
  for (max=copy=i=0; i<count; i++) {
    if ( dots[i]->event == 'b' )
      copy++;
    else
      copy--;
    if ( copy > max )
      max = copy;
  }
  return max;
}

void printNodes(vector<node*>& nodes, ostream& os) {
  vector<node*>::iterator itNode;
  list<bzTag*>::iterator itTag;
  list<bzTag*>* tags;
  vector<eventDot*>* dots;
  eventDot* dota, *dotb;
  int max1, max2, record=0, dotsize, j;

  dots = new vector<eventDot*>;
  for (itNode = nodes.begin(); itNode < nodes.end(); itNode++) {
    tags = (*itNode)->tags;
    for (itTag=tags->begin(); itTag != tags->end(); itTag++) {
      dota = new eventDot;
      dota->event = 'b';
      dota->pos = (*itTag)->b1;
      dotb = new eventDot;
      dotb->event = 'e';
      dotb->pos = (*itTag)->e1;
      dots->push_back(dota);
      dots->push_back(dotb);
    }
    sort(dots->begin(), dots->end(), compar_eventDot);
    max1 = determNumCopy(*dots);

    dotsize = dots->size();
    for (j=0; j<dotsize; j++)
      delete (*dots)[j];
    dots->clear(); // get rid of all elements

    for (itTag=tags->begin(); itTag != tags->end(); itTag++) {
      dota = new eventDot;
      dota->event = 'b';
      dota->pos = (*itTag)->b2;
      dotb = new eventDot;
      dotb->event = 'e';
      dotb->pos = (*itTag)->e2;
      dots->push_back(dota);
      dots->push_back(dotb);
    }
    sort(dots->begin(), dots->end(), compar_eventDot);
    max2 = determNumCopy(*dots);
    os << record++ << ' ' << (*itNode)->chr << ' '
       << (*itNode)->start - fakestarts[init_chrs[(*itNode)->chr]] << ' ' 
       << (*itNode)->end - fakestarts[init_chrs[(*itNode)->chr]] << ' ';  
    os << max1 << ' ' << max2 << endl;     

    dotsize = dots->size();
    for (j=0; j<dotsize; j++)
      delete (*dots)[j];
    dots->clear();
  } // for each node  
  delete dots;
}


void printNodeDetails(vector<node*>& nodes, ostream& os) {
  int i, nodesize=nodes.size();
  int edgesize, tagsize, score, j;
  edge* e;
  list<edge*>* edges;
  list<edge*>::iterator edgeIter;
  bzTag* tag;

  for (i=0; i<nodesize; i++) {
    os << i << ": " << nodes[i]->edges->size() << " : ";
    edges = nodes[i]->edges;
    edgesize = edges->size();
    nodes[i]->edges->sort(compar_edge_x2);
    for (edgeIter=edges->begin(); edgeIter != edges->end(); ++edgeIter) {
    //for (k=0; k<edgesize; k++) {      
      //e = (*(nodes[i]->edges))[k];
      e = (*edgeIter);
      os << e->node_index;
      tagsize = e->tags->size();
      score = 0;
      for (j=0; j<tagsize; j++) {
	tag = (*(e->tags))[j];
	score += (tag->e1 - tag->b1+1 )*tag->pci;
      }
      os << '(' << score << ") ";
    }
    os << endl;
  }
}
//------------ end of displaying procedures ------//
//------------------------------------------------//


//-----< identify the node which the query fragment belongs to >----
//----- binary search ----------------------------------------------
int searchNode(vector<node*>& nodes, int count, long queryFront, long queryEnd) {
  int start, end, middle; 

  start = 0;
  end = count-1;
  while ( start <= end ) {
    middle = (start + end)/2;
    if ( queryFront > nodes[middle]->end ) 
      start = middle+1;
    else if ( queryEnd < nodes[middle]->start )
      end = middle-1;
    else
      return middle;
  }
  return -1;
}

//-----< assign tags to nodes according to sequence1 >---------
void groupTags1(vector<node*>& nodes, vector<bzTag*>* tags) {
  //vector<bzTag*>::iterator itTag;
  int nodeCount, i, tagCount, index;

  nodeCount = nodes.size();
  tagCount = tags->size();
  for (i=0; i<tagCount; i++) {
    index = searchNode(nodes, nodeCount, (*tags)[i]->b1, (*tags)[i]->e1);
    if ( index < 0 ) {
      //cerr << "Not found index for " << tag->b1 << "--" << tag->e1 << endl;
      delete (*tags)[i];
      continue;
    }
    nodes[index]->tags->push_back((*tags)[i]);
    nodes[index]->chr = (*tags)[i]->chr1;
  }
  tags->clear();
}

// overlapped region between a1--b1 and a2--b2
long overlap_length(long a1, long b1, long a2, long b2) {
  if ( a1>b2 || a2>b1 )
    return 0;
  long over_beg = a1 > a2 ? a1 : a2;
  long over_end = b1 < b2 ? b1 : b2;
  return over_end-over_beg+1;
}


int determine_index(vector<node*>& nodes, int index, int nodesize, long beg, long end) {
  long over1, over2, over3;
  // find a good index1: index1-1? index1? index1+1?
  if (index > 0 && nodes[index-1]->chr == nodes[index]->chr)
    over1 = overlap_length(beg, end, nodes[index-1]->start, nodes[index-1]->end);
  else
    over1 = 0;
  over2 = overlap_length(beg, end, nodes[index]->start, nodes[index]->end);
  if (index < nodesize-1 && nodes[index+1]->chr == nodes[index]->chr)
    over3 = overlap_length(beg, end, nodes[index+1]->start, nodes[index+1]->end);
  else
    over3 = 0;
  if ( over1 >= over2 && over1 >= over3)
    index = index-1;
  else if ( over3 >= over2 && over3 >= over1 )
    index = index+1;
  return index;
}

int determine_index2(vector<node*>& nodes, int index2, int nodesize, bzTag* tag, long& split1, long& split2){
  long over1, over2, over3;
  // find a good index2: index2-1? index2? index2+1?
  if (index2 > 0 && nodes[index2-1]->chr == nodes[index2]->chr)
    over1 = overlap_length(tag->b2, tag->e2, nodes[index2-1]->start, nodes[index2-1]->end);
  else
    over1 = 0;
  over2 = overlap_length(tag->b2, tag->e2, nodes[index2]->start, nodes[index2]->end);
  if (index2 < nodesize-1 && nodes[index2+1]->chr == nodes[index2]->chr)
    over3 = overlap_length(tag->b2, tag->e2, nodes[index2+1]->start, nodes[index2+1]->end);
  else
    over3 = 0;

  split1 = split2 = -1;
  if ( over1 >= over2 && over1 >= over3){  // index2-1 longest
    if ( over2 >= MIN_CHAIN )
      split1 = nodes[index2]->start;
    if ( over3 >= MIN_CHAIN ) 
	split2 = nodes[index2+1]->start;
    index2 = index2-1;
  }
  else if ( over3 >= over2 && over3 >= over1 ) { // index2+1 longest
    if ( over1 >= MIN_CHAIN )
      split1 = nodes[index2]->start;
    if ( over2 >= MIN_CHAIN )
      split2 = nodes[index2+1]->start;
    index2 = index2+1;
  }
  else { // index2 longest
    if ( over1 >= MIN_CHAIN )
      split1 = nodes[index2]->start;
    if ( over3 >= MIN_CHAIN )
      split2 = nodes[index2+1]->start;
  }
  return index2;
}

// cutpoint exists on x axis
bzTag* split_tag1(bzTag* orig, long cutpoint) {
  bzTag* newtag;

  newtag = new bzTag(*orig);
  // update
  newtag->b1 = cutpoint;
  orig->e1 = cutpoint-1;
  if ( newtag->strand2 == '+' ) {
    newtag->b2 = newtag->e2 - (newtag->e1 - newtag->b1); // y1 y2 are the same as old
    orig->e2 = orig->e1 - ( orig->b1 - orig->b2 );
  }
  else {
    newtag->e2 = newtag->b2 - (newtag->b1 - newtag->e1);
    orig->b2 = orig->b1 - ( orig->e1 - orig->e2 );
  }
  return newtag;
}

// cutpoint exists on y axis
bzTag* split_tag2(bzTag* orig, long cutpoint) {
  bzTag* newtag;

  newtag = new bzTag(*orig);
  // update
  newtag->b2 = cutpoint;
  orig->e2 = cutpoint-1;
  if ( newtag->strand2 == '+' ) {
    newtag->b1 = newtag->e1 - (newtag->e2 - newtag->b2); // y1 y2 are the same as old
    orig->e1 = orig->e2 - ( orig->b2 - orig->b1 );
  }
  else {
    newtag->e1 = newtag->b1 - (newtag->b2 - newtag->e2);
    orig->b1 = orig->b2 - ( orig->e2 - orig->e1 );
  }
  return newtag;
}

// no splitting
int determine_node(vector<node*>& nodes, int nodesize, long beg, long end) {
  int index = searchNode(nodes, nodesize, beg, end);
  if ( index < 0 ) {
    //cerr << "index is not found for : " << beg << "--" << end << endl;
    return index;
  }
  index = determine_index(nodes, index, nodesize, beg, end);
  return index;
}

//-----< assume nodes are ordered by start >--------------
// clear tags after done
//-----< tags may be split >------------------------------
void groupTags2(vector<node*>& nodes, vector<bzTag*>* tags) {
  int index2, nodesize=nodes.size(), tagsize = tags->size(), i;
  long split1, split2;
  bzTag* newtag, *tag;
  int index;
  
  for (i=0; i<tagsize; i++) {
    tag = (*tags)[i];
    index2 = searchNode(nodes, nodesize, tag->b2, tag->e2);
    if ( index2 < 0 ) {
      //cerr << "index2 is not found for: " << tag->b2 << "--" << tag->e2 << endl;
      delete tag;
      continue;
    }
    // decide whether to split 
    index2 = determine_index2(nodes, index2, nodesize, tag, split1, split2);
    if ( split2 > 0 ){
      newtag = split_tag2(tag, split2);
      index = determine_node(nodes, nodesize, newtag->b1, newtag->e1);
      if ( index < 0 ) {
	// cerr << "index(1) is not found for: " << newtag->b1 << ' ' << newtag->e1 << endl;
	delete newtag;
      }
      else
	nodes[index]->tags->push_back(newtag);
    } 
    if ( split1 > 0 ) {
      newtag = split_tag2(tag, split1);
      index = determine_node(nodes, nodesize, newtag->b1, newtag->e1);
      if ( index < 0 ) {
	//cerr << "index(2) is not found for: " << newtag->b1 << ' ' << newtag->e1 << endl;
	delete newtag;
      }
      else
	nodes[index]->tags->push_back(newtag);
    }
    
    index = determine_node(nodes, nodesize, tag->b1, tag->e1);
    if ( index < 0 ) {
      //cerr << "index(3) is not found for: " << tag->b1 << "--" << tag->e1 << endl;
      delete tag;
    }
    else
      nodes[index]->tags->push_back(tag);
  }
  tags->clear();
}

//-----------------------------------------------------
//
//  Get rid of small overlapping
//  (replace with non-linear optimization??)
//
//-----------------------------------------------------
// tag1 is longer than tag2, both are positive strand
// return true: remove tag2
bool trim_pos_pos(bzTag* tag1, bzTag* tag2) {

  if ( (tag1->b2 - tag1->b1 ) - (tag2->b2 - tag2->b1) > CLOSE_THRESHOLD)
    return false;
  if ( (tag2->b2 - tag2->b1 ) - (tag1->b2 - tag1->b1) > CLOSE_THRESHOLD)
    return false;

  long over_beg, over_end;
  over_beg = tag1->b1 > tag2->b1 ? tag1->b1 : tag2->b1;
  over_end = tag1->e1 < tag2->e1 ? tag1->e1 : tag2->e1;
  if ( ( over_end-over_beg+1 )*100/( tag2->e1 - tag2->b1 + 1 ) > 80 )
    return true;
  over_beg = tag1->b2 > tag2->b2 ? tag1->b2 : tag2->b2;
  over_end = tag1->e2 < tag2->e2 ? tag1->e2 : tag2->e2;
  if ( ( over_end-over_beg+1 )*100/( tag2->e2 - tag2->b2 + 1 ) > 80 )
    return true;

  if ( (tag2->e1 > tag1->e1) && (tag2->b1 <= tag1->e1) ) {
    tag2->b1 = tag1->e1+1;
    tag2->b2 = tag2->b1 - (tag2->e1 - tag2->e2);
  }
  else if ( (tag2->b1 < tag1->b1) && (tag2->e1 >= tag1->b1) ) {
    tag2->e1 = tag1->b1-1;
    tag2->e2 = tag2->e1 - (tag2->b1 - tag2->b2);
  }
  if ( (tag2->e2 > tag1->e2) && (tag2->b2 <= tag1->e2) ) {
    tag2->b2 = tag1->e2+1;
    tag2->b1 = tag2->b2 - (tag2->e2 - tag2->e1);
  }
  else if ( (tag2->b2 < tag1->b2) && (tag2->e2 >= tag1->b2) ) {
    tag2->e2 = tag1->b2-1;
    tag2->e1 = tag2->e2 - (tag2->b2 - tag2->b1);
  }
  return false;
}

// tag1 is longer than tag2, both are neg strand
// but tag1, tag2's fileds x1, x2, y1, y2 are virtual values on pos strands
// return true: remove tag2
bool trim_neg_neg(bzTag* tag1, bzTag* tag2) {
  if ( (tag1->b1 + tag1->e2) - (tag2->b1 + tag2->e2) > CLOSE_THRESHOLD)
    return false;
  if ( (tag2->b1 + tag2->e2) - (tag1->b1 + tag1->e2) > CLOSE_THRESHOLD)
    return false;
  
  long over_beg, over_end;
  over_beg = tag1->b1 > tag2->b1 ? tag1->b1 : tag2->b1;
  over_end = tag1->e1 < tag2->e1 ? tag1->e1 : tag2->e1;
  if ( ( over_end-over_beg+1 )*100/( tag2->e1 - tag2->b1 + 1 ) > 80 )
    return true;
  over_beg = tag1->b2 > tag2->b2 ? tag1->b2 : tag2->b2;
  over_end = tag1->e2 < tag2->e2 ? tag1->e2 : tag2->e2;
  if ( ( over_end-over_beg+1 )*100/( tag2->e2 - tag2->b2 + 1 ) > 80 )
    return true;

  if ( (tag2->e1 > tag1->e1) && (tag2->b1 <= tag1->e1) ) {
    tag2->b1 = tag1->e1+1;
    tag2->e2 = tag2->e1 - (tag2->b1 - tag2->b2);
  }
  else if ( (tag2->b1 < tag1->b1) && (tag2->e1 >= tag1->b1) ) {
    tag2->e1 = tag1->b1-1;
    tag2->b2 = tag2->b1 - ( tag2->e1 - tag2->e2 );
  }						
  if ( (tag2->e2 > tag1->e2) && (tag2->b2 <= tag1->e2) ) {
    tag2->b2 = tag1->e2+1; 
    tag2->e1 = tag2->e2 - (tag2->b2 - tag2->b1);
  }
  else if ( (tag2->b2 < tag1->b2) && (tag2->e2 < tag1->b2) ) {
    tag2->e2 = tag1->b2-1;
    tag2->b1 = tag2->b2 - (tag2->e2 - tag2->e1);
  }
  return false;
}

// tag1 is longer than tag2, tag1 is pos strand, tag2 is neg strand
// tag2's fields x1, x2, y1, y2 are virtual values on pos strand
// return true: remove tag2
bool trim_pos_neg(bzTag* tag1, bzTag* tag2){
  //int b  = tag2->b1 + tag2->e2;
  //int b1 = tag1->e1 + tag1->e2;
  
  //if ( b - b1  > CLOSE_THRESHOLD )
  if ( tag2->b1 - tag1->e1 + tag2->e2 - tag1->e2 > CLOSE_THRESHOLD )
    return false;
  //int b2 = tag1->b1 + tag1->b2;
  if ( tag1->b1 - tag2->b1 + tag1->b2 - tag2->e2 > CLOSE_THRESHOLD )
    //if ( b2 - b > CLOSE_THRESHOLD )
    return false;
  long  e = tag1->b2 - tag1->b1;
  long e1 = tag2->e2 - tag2->b1;
  if ( e - e1 > CLOSE_THRESHOLD )
    return false;
  long e2 = tag2->b2 - tag2->e1;
  if ( e2 - e > CLOSE_THRESHOLD )
    return false;
  
  long over_beg, over_end;
  over_beg = tag1->b1 > tag2->b1 ? tag1->b1 : tag2->b1;
  over_end = tag1->e1 < tag2->e1 ? tag1->e1 : tag2->e1;
  if ( (over_end-over_beg+1)*100/(tag2->e1-tag2->b1+1) > 80 )
    return true;
  over_beg = tag1->b2 > tag2->b2 ? tag1->b2 : tag2->b2;
  over_end = tag1->e2 < tag2->e2 ? tag1->e2 : tag2->e2;
  if ( (over_end-over_beg+1)*100/(tag2->e2-tag2->b2+1) > 80 )
    return true;

  if ( (tag2->e1 > tag1->e1) && (tag2->b1 <= tag1->e1) ) {
    tag2->b1 = tag1->e1+1;
    tag2->e2 = tag2->e1 - (tag2->b1 - tag2->b2);
  }
  else if ( (tag2->b1 < tag1->b1) && (tag2->e1 >= tag1->b1) ) {
    tag2->e1 = tag1->b1-1;
    tag2->b2 = tag2->b1 - (tag2->e1 - tag2->e2);
  }
  if ( (tag2->e2 > tag1->e2) && (tag2->b2 <= tag1->e2) ) {
    tag2->b2 = tag1->e2+1;
    tag2->e1 = tag2->e2 - (tag2->b2 - tag2->b1);
  }
  else if ( (tag2->b2 < tag1->b2) && (tag2->e2 < tag1->b2) ) {
    tag2->e2 = tag1->b2-1;
    tag2->b1 = tag2->b2 - (tag2->e2 - tag2->e1);
  }
  return false;
}

// tag1 is longer than tag2, tag1 is neg strand, tag2 is pos strand
// tag1's fields x1, x2, y1, y2 are virtual values on pos strand
// return true: remove tag2
bool trim_neg_pos(bzTag* tag1, bzTag* tag2) {
  long b  = tag2->b2 - tag2->b1;
  long b1 = tag1->e2 - tag1->b1;
  if ( b - b1 > CLOSE_THRESHOLD )
    return false;
  long b2 = tag1->b2 - tag1->e1;
  if ( b2 - b > CLOSE_THRESHOLD )
    return false;
  //int e  = tag1->b1 + tag1->e2; // potential overflow
  //int e1 = tag2->b1 + tag2->b2;
  //if ( e1 - e > CLOSE_THRESHOLD )
  if ( tag2->b1 - tag1->b1 + tag2->b2 - tag1->e2 > CLOSE_THRESHOLD )
    return false;
  //int e2 = tag2->e1 + tag2->e2;
  //if ( e - e2 > CLOSE_THRESHOLD )
  if ( tag1->b1 - tag2->e1 + tag1->e2 - tag2->e2 > CLOSE_THRESHOLD )
    return false;

  long over_beg, over_end;
  over_beg = tag1->b1 > tag2->b1 ? tag1->b1 : tag2->b1;
  over_end = tag1->e1 < tag2->e1 ? tag1->e1 : tag2->e1;
  if ( (over_end-over_beg+1)*100/(tag2->e1-tag2->b1+1) > 80 )
    return true;
  over_beg = tag1->b2 > tag2->b2 ? tag1->b2 : tag2->b2;
  over_end = tag1->e2 < tag2->e2 ? tag1->e2 : tag2->e2;
  if ( (over_end-over_beg+1)*100/(tag2->e2-tag2->b2+1) > 80 )
    return true;

  if ( (tag2->e1 > tag1->e1) && (tag2->b1 <= tag1->e1) ) {
    tag2->b1 = tag1->e1+1;
    tag2->b2 =  tag2->b1 - (tag2->e1 - tag2->e2);
  }
  else if ( (tag2->b1 < tag1->b1) && (tag2->e1 >= tag1->b1) ) {
    tag2->e1 = tag1->b1-1;
    tag2->e2 = tag2->e1 - (tag2->b1 - tag2->b2);
  }
  if ( (tag2->e2 > tag1->e2) && (tag2->b2 <= tag1->e2) ){
    tag2->b2 = tag1->e2+1;
    tag2->b1 =  tag2->b2 - (tag2->e2 - tag2->e1);
  }
  else if ( (tag2->b2 < tag1->b2) && (tag2->e2 >= tag1->b2) ) {
    tag2->e2 = tag1->b2-1;
    tag2->e1 =  tag2->e2 - (tag2->b2 - tag2->b1);
  }
  return false;
}

void trimTags(vector<node*>& nodes) {
  int nodeArrSize=nodes.size(), i, tagArrSize, j;
  list<bzTag*>* tags;
  list<bzTag*>::iterator tagIt1, tagIt2, tagIt1next, tagIt2next, cutIt1, cutIt2;
  list<bzTag*>::reverse_iterator rTagIt1, rTagIt1next;
  bzTag *t1, *t2;
  bool ret;
  int eraseCut2, eraseCut1, cut2isTag1, cut1, cut2, index1, index2;
  char* eraseArr;

  for (i=0; i<nodeArrSize; i++) {
    tags = nodes[i]->tags;
    tagArrSize = tags->size();
    if ( tagArrSize == 0 )
      continue;
    tags->sort(compar_tag_x2);
    eraseArr = (char*)malloc(tagArrSize*sizeof(char));
    memset(eraseArr, 'n', tagArrSize);
    //for ( index1 = 0; index1 < tagArrSize; index1++) {
    for (index1=0, tagIt1 = tags->begin(); index1 < tagArrSize; ++index1, ++tagIt1) {
      if ( eraseArr[index1] == 'y' )
	continue;
      t1 = *tagIt1;
      tagIt2 = tagIt1;
      tagIt2++;
      for ( index2 = index1+1; index2 < tagArrSize; ++index2, ++tagIt2) {
	if ( eraseArr[index2] == 'y' )
	  continue;
	t2 = *tagIt2;
	if ( t2->b2 > t1->e2 + CLOSE_THRESHOLD ) // too far
	  break;
	if ( t2->b1 > t1->e1 && t2->b2 > t1->e2 )
	  continue;
	if ( t2->b1 > t1->e1 && t2->e2 < t1->b2 )
	  continue;
	if ( t2->e1 < t1->b1 && t2->b2 > t1->e2 )
	  continue;
	if ( t2->e1 < t1->b1 && t2->e2 < t1->b2 )
	  continue;
	
	eraseCut1 = eraseCut2 = cut2isTag1 = 0;
	if ( (t1->e1 - t1->b1) > (t2->e1 - t2->b1) ) { // need to cut tags[index2]
	  cut1 = index1;
	  cut2 = index2;
	  cutIt1 = tagIt1;
	  cutIt2 = tagIt2;
	}
	else {
	  cut1 = index2;
	  cut2 = index1;
	  cutIt1 = tagIt2;
	  cutIt2 = tagIt1;
	  cut2isTag1 = 1;
	}
	if ( (*cutIt1)->strand2 == '+' ){
	  if ( (*cutIt2)->strand2 == '+')
	    ret = trim_pos_pos(*cutIt1, *cutIt2);
	  else
	    ret = trim_pos_neg(*cutIt1, *cutIt2);
	}
	else {
	  if ( (*cutIt2)->strand2 == '+')
	    ret = trim_neg_pos(*cutIt1, *cutIt2);
	  else
	    ret = trim_neg_neg(*cutIt1, *cutIt2);
	}
	if ( ret == true ) {
	  eraseArr[cut2] = 'y';
	  eraseCut2 = 1;
	}
	else if ( (*cutIt2)->e1 - (*cutIt2)->b1 < SML_MIN_LENGTH ) {
	  eraseArr[cut2] = 'y';
	  eraseCut2 = 1;
	}
	if ( (*cutIt1)->e1 - (*cutIt1)->b1 < SML_MIN_LENGTH ) {
	  eraseArr[cut1] = 'y';
	  eraseCut1 = 1;
	}
	if ( (eraseArr[cut2]=='y' && cut2isTag1==1) || (eraseArr[cut1]=='y' && cut2isTag1==0 ) )
	//if ( (eraseCut2==1 && cut2isTag1==1) || (eraseCut1==1 && cut2isTag1==0) )
	  break;

      }// for pair
    }// each tag
    tagIt1 = tags->end();
    tagIt1--;
    for (j = tagArrSize-1; j >= 0; j--, tagIt1= tagIt1next ) {
      tagIt1next = tagIt1;
      tagIt1next--;
      if ( eraseArr[j] == 'y' ) {
	delete *tagIt1;
	tags->erase( tagIt1 );
      }
    }
    //sort( tags.begin(), tags.end(), compar_tag_x1 );
    free(eraseArr);
  } // for each node

  for (i=nodeArrSize-1; i>=0; i--)
    if ( nodes[i]->tags->size() == 0 ) {
      delete nodes[i]->tags;
      delete nodes[i]->edges;
      delete nodes[i];
      nodes.erase( nodes.begin() + i ); // remove node* from vector
    }
}


// based on the threshold, this chain function chains close tags,
// over longer distance
vector< vector<bzTag*>* >* chainingTags(list<bzTag*>* tags, int threshold) {
  vector< vector<bzTag*>* >* chains = new vector< vector<bzTag*>* >;
  vector<bzTag*>* newchain;
  vector<bzTag*> candidates;
  list<bzTag*>::iterator tagIt, tagIt2, tagIt3;
  bzTag* curr;
  int i, j, k, count;
  char* markArr;

  tags->sort(compar_tag_x2);
  count = tags->size();
  markArr = (char*)malloc(count*sizeof(char));
  memset(markArr, 'n', count);
  //for (i=0; i<count; i++) { // for each tag
  for (tagIt=tags->begin(), i=0; tagIt != tags->end(); ++tagIt, ++i ){
    if ( markArr[i] == 'y' ) // already a part of chain
      continue;
    newchain = new vector<bzTag*>;
    newchain->push_back(*tagIt);
    markArr[i] = 'y';
    chains->push_back(newchain);
    curr = *tagIt; // current tag
    candidates.clear();
    tagIt2=tagIt;
    tagIt2++;
    for ( j=i+1; j<=count; ++tagIt2, ++j) { // a single chain, for each next tag
      if ( j<count ) {
	if ( markArr[j] == 'y' )   // already a part of chain
	  continue;
	if ( (*tagIt2)->strand2 != curr->strand2 ) // different orientation
	  continue;
	if ( (*tagIt2)->b2 < curr->e2 ) //regardless of strand
	  continue;  
	if ( (((*tagIt2)->b1 < curr->e1 ) && (curr->strand2 == '+' ))
	     || (((*tagIt2)->e1 > curr->b1 ) && (curr->strand2 == '-' )) )
	  continue;   // overlapped on x
      }
      if (  j==count || (*tagIt2)->b2 - curr->e2 > threshold  ) { // regardless of strand
	if ( candidates.size() == 0 )
	  break;
	// otherwise need to find the best candidate
	origY2 = curr->e2;
	if ( curr->strand2 == '+' ) {
	  origY1 = curr->e1;
	  sort(candidates.begin(), candidates.end(), compar_tags_distancePos);
	}
	else { // '-'
	  origX1 = curr->b1;
	  sort(candidates.begin(), candidates.end(), compar_tags_distanceNeg);
	}
	curr = candidates[0]; // the closest one
	newchain->push_back( curr );
	tagIt3 = tagIt;
	tagIt3++;
	for (k=i+1; k<count; ++tagIt3, ++k)
	  if ( (*tagIt3)->b1 == candidates[0]->b1 && (*tagIt3)->b2 == candidates[0]->b2 )
	    break;
	markArr[k] = 'y';
	candidates.clear();   // does it destroy or just remove elements
	j = k;   // next must be fron next of k
	tagIt2 = tagIt3;
      }
      else // distance < threshold
	candidates.push_back( (*tagIt2) );
    }
  }// for each tag
  // sort tags in each chain
  free(markArr);
  count = chains->size();
  for (i=0; i<count; i++) {
    newchain = (*chains)[i];
    sort(newchain->begin(), newchain->end(), compar_tag_x1);
  }

  sort( chains->begin(), chains->end(), compar_chain_x2 );
  return chains;
}

void delete_chains(vector< vector<bzTag*>* >* chains) {
  vector<bzTag*>* chain;
  int chainSize=chains->size(), j;

  for (j=0; j<chainSize; j++) {
    chain = (*chains)[j];
    chain->clear();
    delete (*chains)[j];
  }
  chains->clear();
  delete chains;
}

// remove tags whose chains below MIN_CHAIN threshold
void remove_small_chains(vector<node*>& nodes) {
  int nodeSize = nodes.size(), k,j, chainSize, chainTagSize, i, nodeTagSize, m;
  unsigned int max_tag_size=0;
  char *tagMarkArr;
  vector< vector<bzTag*>* >* chains;
  vector<bzTag*>* chainTags;
  list<bzTag*>* nodeTags;
  list<bzTag*>::iterator tagIt, tagItprev, tagItnext;
  list<bzTag*>::reverse_iterator rTagIt, rTagItnext;

  for (k=0; k<nodeSize; k++)
    if ( max_tag_size < nodes[k]->tags->size() )
      max_tag_size = nodes[k]->tags->size();

  tagMarkArr = (char*)malloc(max_tag_size*sizeof(char));

  for (k=0; k<nodeSize; k++) {
    if ( nodes[k] == NULL )
      continue;
    nodeTags = nodes[k]->tags;
    nodeTagSize = nodeTags->size();
    memset(tagMarkArr, 'n', nodeTagSize);
    chains =  chainingTags( nodes[k]->tags, NODE_THRESHOLD);
    chainSize = chains->size();
    for (j=0; j<chainSize; j++) {
      chainTags = (*chains)[j];
      chainTagSize = chainTags->size();
      if ( (*chainTags)[chainTagSize-1]->e1 - (*chainTags)[0]->b1 < MIN_CHAIN ) {
	// get rid of these tags from node's tag vector
	for (m=0; m<chainTagSize; m++) {
	  for (i=0, tagIt=nodeTags->begin(); i<nodeTagSize; ++tagIt, i++)
	    if ( (*tagIt)->b1 == (*chainTags)[m]->b1 
		 && (*tagIt)->b2 == (*chainTags)[m]->b2 )
	      break;
	  if ( i==nodeTagSize ) {
	    //cerr << "not found: " << m << endl;
	    continue;
	  }
	  tagMarkArr[i] = 'y';
	}// for each chain tag
      }// if
    } // for each chain in each node
    delete_chains(chains);
    tagIt = nodeTags->end();
    tagIt--;
    for (j=nodeTagSize-1; j>=0; j--, tagIt=tagItnext) {
      tagItnext = tagIt;
      tagItnext--;
      if ( tagMarkArr[j] == 'y' ) {
	delete *tagIt;
	nodeTags->erase(tagIt);
      }
    }
  }// for each node
  free(tagMarkArr);
}

// but the edges object is not deleted yet. edges becomes empty
void clean_edges(list< edge* >* edges) {
  list<edge*>::iterator edgeIt;
  edge* e;

  if ( edges->size() > 0 )
    for (edgeIt=edges->begin(); edgeIt != edges->end(); ++edgeIt) {
      e = *edgeIt;
      e->tags->clear();
      delete e->tags;
      delete e;
    }
  edges->clear();
}


void maintenance_nodes(vector<node*>& nodes) {
  int i, nodesize = nodes.size();
  char* eraseArr = (char*)malloc(nodesize*sizeof(char));
 
  memset(eraseArr, 'n', nodesize);
  for (i=0; i<nodesize; i++) {
    if ( nodes[i] == NULL )
      eraseArr[i] = 'y';
    else if ( nodes[i]->tags->empty() )
      eraseArr[i] = 'y';
    else {
      nodes[i]->tags->sort(compar_tag_x1);
      if ( nodes[i]->start < (*(nodes[i]->tags->begin()))->b1 )
	nodes[i]->start = (*(nodes[i]->tags->begin()))->b1;
      if ( nodes[i]->end > (nodes[i]->tags->back())->e1 )
	nodes[i]->end   = (nodes[i]->tags->back())->e1;
    }
  }
  for (i=nodesize-1; i>=0; i--)
    if ( eraseArr[i] == 'y' ) {
      if ( nodes[i] != NULL ) {
	clean_edges(nodes[i]->edges);
	delete nodes[i]->edges;
	delete nodes[i]->tags;
	delete nodes[i];
      }
      nodes.erase( nodes.begin() + i );
    }
  free(eraseArr);
}

// chains belong to the same node, need to determne whether the node 
// need to break, and the position of breakpoint.
// determine the pair of chains which overlap most on x2, then decide to cut
long getBreakPoint(vector< vector<bzTag*>* >* chains) {
  int *currEndsIndices, chainArrSize=chains->size();
  long* currEnds;
  long maxOver, currOver, left_range, right_range;
  int maxi=0, maxj=0, index;
  vector<long> starts, ends;

  // sort each chain

  currEnds = (long*)malloc(chainArrSize*sizeof(long));
  currEndsIndices = (int*)malloc(chainArrSize*sizeof(int));
  currEnds[0] = (*((*chains)[0]))[0]->b2 - 1;
  currEndsIndices[0] = -1;
  starts.push_back( (*((*chains)[0]))[0]->b1 );
  ends.push_back( (*(((*chains)[0])->end()-1))->e1);

  for (index=1; index<chainArrSize; ++index) {
    currEnds[index] = ( (*(((*chains)[index-1])->end()-1))->e2 > currEnds[index-1] ? (*(((*chains)[index-1])->end()-1))->e2 : currEnds[index-1] );
    if ( currEnds[index] == currEnds[index-1] )
      currEndsIndices[index] = currEndsIndices[index-1];
    else
      currEndsIndices[index] = index - 1;
    starts.push_back( (*(((*chains)[index])->begin()))->b1 );
    ends.push_back( (*(((*chains)[index])->end()-1))->e1);
  }

  maxOver = 0;
  for (index = 1; index < chainArrSize; index++) {
    if ( currEnds[index] < (*(((*chains)[index])->end()-1))->e2 )
      currOver = currEnds[index] - (*(((*chains)[index])->begin()))->b2 + 1;
    else
      currOver = (*(((*chains)[index])->end()-1))->e2 - (*(((*chains)[index])->begin()))->b2 + 1;
    if ( currOver > maxOver ) {
      maxOver = currOver;
      maxi = currEndsIndices[index];
      maxj = index;
    }
  }

  free(currEndsIndices);
  free(currEnds);

  if ( maxOver < MAX_SPLIT_OVERLAP)
    return 0;  // no need to cut since no overlap
   
  long maxi_x1, maxj_x1, maxj_y1, maxi_y1, breakpoint, maxi_x2, maxj_x2;
  //long diff1;
  maxi_x1 = (*(((*chains)[maxi])->begin()))->b1;
  maxi_y1 = (*(((*chains)[maxi])->end()-1))->e1;
  maxi_x2 = (*(((*chains)[maxi])->begin()))->b2;
  maxj_x1 = (*(((*chains)[maxj])->begin()))->b1;
  maxj_y1 = (*(((*chains)[maxj])->end()-1))->e1;
  maxj_x2 = (*(((*chains)[maxj])->begin()))->b2;
  int middle_point, half_distance, swap_tmp;
  long temp;
  // point j locate left to i, need to swap
  if ( ( maxj_x2 - maxi_x2 > maxj_x1 - maxi_x1 && maxj_x1 - maxi_x1 >= 0 )
       || ( maxj_x2 - maxi_x2 <= 0 && maxj_x2 - maxi_x2 >  maxj_x1 - maxi_x1 )
       || ( maxj_x1 - maxi_x1 <= 0 && maxj_x2 - maxi_x2 >= 0) ) {
    swap_tmp = maxj;
    maxj = maxi;
    maxi = swap_tmp;
    maxi_x1 = (*(((*chains)[maxi])->begin()))->b1;
    maxi_y1 = (*(((*chains)[maxi])->end()-1))->e1;
    maxi_x2 = (*(((*chains)[maxi])->begin()))->b2;
    maxj_x1 = (*(((*chains)[maxj])->begin()))->b1;
    maxj_y1 = (*(((*chains)[maxj])->end()-1))->e1;
    maxj_x2 = (*(((*chains)[maxj])->begin()))->b2;
  }

  // in shape of 
  // if nonoverlap on x
  if ( maxj_x1 > maxi_y1 ) {
    //breakpoint = maxj_x1/2 + maxi_y1/2;  //****** error prone *****//
    left_range = maxi_x1/2 + maxi_y1/2;
    right_range = maxj_x1/2 + maxj_y1/2;
    breakpoint = graph_cutpoint(starts, ends, left_range, right_range);
  }
  else { // overlap
    if ( maxj_x2 <= maxi_x2 ) {
	middle_point = maxj_x1 + ( maxi_x2 - maxj_x2 );
	half_distance = (middle_point - maxi_x1)/2;
    }
    else { // maxj_x2 > maxi_x2
	middle_point = maxj_x1;
	half_distance = (maxj_x1 - maxi_x1 - (maxj_x2 - maxi_x2 ))/2;
    }
    left_range = middle_point - half_distance;
    right_range = middle_point + half_distance;

    /*
    diff1 = maxj_x1 - maxi_x1;
    left_range = maxi_x1 + diff1/2;
    //right_range = maxj_x1 + diff1/2 < maxi_y1 ? maxj_x1 + diff1/2 : maxi_y1;
    right_range = maxj_x1 + diff1/2 > maxi_y1 ? maxj_x1 + diff1/2 : maxi_y1;
    */

    if ( left_range > right_range) {
      temp = left_range;
      left_range = right_range;
      right_range = temp;
    }
    breakpoint = graph_cutpoint(starts, ends, left_range, right_range);
    if ( breakpoint - maxi_x1 < CUT_THRESHOLD)
      return 0;
    if ( maxj_y1 - breakpoint < CUT_THRESHOLD)
      return 0;
    /*
    diff2 = maxj_y1 - maxi_y1;
    if ( diff2 < 0 )
      diff2 = -diff2;
    if ( diff1 > diff2 ) {
      breakpoint = maxj_x1;
      if ( breakpoint - maxi_x1 < CUT_THRESHOLD ) // is it short repeat region??
	return 0;
    } 
    else if ( maxj_y1 > maxi_y1 ) {
      breakpoint = maxi_y1 + 1;
      if ( maxj_y1 - breakpoint < CUT_THRESHOLD ) 
	return 0;
    }
    else {
      breakpoint = maxj_y1 + 1;
      if ( maxi_y1 - breakpoint < CUT_THRESHOLD )
	return 0;
    }
    */
  }
  return breakpoint;
}

void construct_graph_node(vector<node*>& nodes, int currPos) {
  node* n = nodes[currPos];
  bzTag* t;
  list<edge*>* edges;
  list<edge*>::iterator edgeIt;
  edge *ne;
  list<bzTag*>* tags;
  list<bzTag*>::iterator tagIt;
  int index1, index2, nodesize=nodes.size();
  
  tags = n->tags;
  for (tagIt=tags->begin(); tagIt!=tags->end(); ++tagIt) {
    t = *tagIt;
    index1 = determine_node(nodes, nodesize, t->b1, t->e1);
    if ( index1 < 0 ) {
      //cerr << "Not found index1: " << currPos << ' ' << t->b1 << ' ' << t->e1 << endl;
      continue;
    }
    if ( index1 != currPos ) {
      //cerr << "Not consistent: " << currPos << ' ' << index1 << endl;
      continue;
    }
    //index2 = searchNode(nodes, nodesize, t->b2, t->e2);
    index2 = determine_node(nodes, nodesize, t->b2, t->e2);
    if ( index2 < 0 ) {
      //cerr << "Not found index2: " << currPos << ' ' << t->b2 << ' ' << t->e2 << endl;
      continue;
    }
    edges = n->edges;
    if ( edges->size() > 0 ) {
      for (edgeIt=edges->begin(); edgeIt != edges->end(); ++edgeIt)
	if ( (*edgeIt)->node_index == index2 )
	  break;
      if ( edgeIt != edges->end() ) { // edge found
	(*edgeIt)->tags->push_back(t);
	(*edgeIt)->score += t->pci*(t->e1-t->b1+1); 
      }
    }
    if ( edges->size() == 0 || edgeIt == edges->end() ) { // edge not found, maybe 0
      ne = new edge;
      ne->node_index = index2;
      ne->score = t->pci*(t->e1-t->b1+1);
      ne->tags = new vector<bzTag*>;
      ne->tags->push_back(t);
      n->edges->push_back(ne);
    }
  }//for each tag in a node

  for (edgeIt=n->edges->begin(); edgeIt != n->edges->end(); ++edgeIt)
    sort( (*edgeIt)->tags->begin(), (*edgeIt)->tags->end(), compar_tag_x1);

  n->edges->sort(compar_edge_x1);
}


void cutNode(vector<node*>& nodes, int currPos, long breakpoint, int cutEdges=1) {
  list<bzTag*>* tags = nodes[currPos]->tags;
  list<bzTag*>::iterator tagIt, tagItnext;
  list<bzTag*>::reverse_iterator rTagIt, rTagItnext;
  bzTag* newtag;
  node* newnode, *n = nodes[currPos];
  char* eraseArr;
  int tagArrSize, i;

  tagArrSize = tags->size();
  tags->sort(compar_tag_x1);
  // create the new node
  newnode = new node;
  newnode->start = breakpoint;
  newnode->end = n->end;
  newnode->chr = n->chr;
  newnode->tags = new list<bzTag*>;
  newnode->edges = new list<edge*>;
  nodes.push_back(newnode); // at the end
  n->end = breakpoint-1;    // n is the curr node
  // split tags
  eraseArr = (char*)malloc(tagArrSize*sizeof(char));
  memset(eraseArr, 'n', tagArrSize);
  for ( i=0, tagIt=tags->begin(); i<tagArrSize; ++tagIt, i++ ) {
    if ( (*tagIt)->b1 >= breakpoint ) { // mv to new node
      newnode->tags->push_back(*tagIt);
      eraseArr[i] = 'y';
    }
    else if ( (*tagIt)->b1 < breakpoint && (*tagIt)->e1 > breakpoint ) { // update, create new
      newtag = split_tag1(*tagIt, breakpoint);

      if ( newtag->e1 - newtag->b1 >= MIN_LENGTH ) {
	newnode->tags->push_back(newtag);
      }
      else
	delete newtag;
      if ( (*tagIt)->e1 - (*tagIt)->b1 < MIN_LENGTH ) {
	delete *tagIt;
	eraseArr[i] = 'y';
      }
    }
  }
  
  tagIt = tags->end();
  tagIt--;
  for (i=tagArrSize-1; i>=0; i--, tagIt=tagItnext) {
    tagItnext = tagIt;
    tagItnext--;
    if ( eraseArr[i] == 'y' ) {
      //delete *tagIt;
      tags->erase( tagIt );
    }
  }

  free(eraseArr);

  if ( nodes[currPos]->edges->empty() )
    return;

  // split edges
  
  if (cutEdges == 1 ) {
    clean_edges(nodes[currPos]->edges);
    clean_edges((*(nodes.end()-1))->edges);
    // do sorting since construct_graph_node assume nodes are in order
    //sort(nodes.begin(), nodes.end(), compar_node_start);
    //construct_graph_node(nodes, currPos);
    //construct_graph_node(nodes, currPos+1); // it must be right after currPos
  }
}


//-----< split a node, and its tags >-------------------
// sort according to x2 (starting position of sequence 2
// prepare an extra array to record currEnd of sequence 2
// find the pair of lines overlap most on y, split on x
// recursively call each of the splitted node
void splitNode(vector<node*>& nodes, int currPos) {
  vector< vector<bzTag*>* >* chains;
  long breakpoint;

  if ( currPos == 303 )
    printf("hello");

  if ( nodes[currPos]->tags->empty() )
    return;

  //cout << currPos << endl;
  chains = chainingTags(nodes[currPos]->tags, GAP_THRESHOLD); // small values for repeat

  breakpoint = getBreakPoint(chains);
  delete_chains(chains);
  if ( breakpoint == 0 )
    return;

  if ( breakpoint < nodes[currPos]->start || breakpoint > nodes[currPos]->end)
    cerr << "breakpoint hello\n";

  cutNode(nodes, currPos, breakpoint);

  splitNode(nodes, currPos);  // the updated curr element
  splitNode(nodes, nodes.size()-1); // the new inserted last element
}


void splitNodes(vector<node*>& nodes) {
  int k, nodesize = nodes.size();
  for (k=0; k<nodesize; k++)
    splitNode(nodes, k);
}

bool overlap(long a1, long a2, long b1, long b2, int threshold, int looser=0) {
  if ( a1 - b2 > -threshold )
    return false;
  if ( b1 - a2 > -threshold )
    return false;
  if ( looser == 0 )
    return true;

  long length1, length2, over_beg, over_end, over_len;
  over_beg = a1 > b1 ? a1 : b1;
  over_end = a2 < b2 ? a2 : b2;
  over_len = over_end - over_beg + 1;
  length1 = (a2-a1+1);
  length2 = (b2-b1+1);

  if ( over_len*100/length1 < 50 && over_len*100/length2 < 50 )
    return false;
  return true;
}

// test if two chains overlap on Y
bool chain_overlap(vector< bzTag* >* chain1, vector< bzTag* >* chain2 ) {
  long a1, a2, b1, b2;
  if ( (*chain1)[0]->strand2 == '+' ) {
    a1 = (*(chain1->begin()))->b2;
    a2 = (*(chain1->end()-1))->e2;
  }
  else {
    a1 = (*(chain1->end()-1))->b2;
    a2 = (*(chain1->begin()))->e2;
  }
  if ( (*chain2)[0]->strand2 == '+' ) {
    b1 = (*(chain2->begin()))->b2;
    b2 = (*(chain2->end()-1))->e2;
  }
  else {
    b1 = (*(chain2->end()-1))->b2;
    b2 = (*(chain2->begin()))->e2;
  }
  return overlap(a1, a2, b1, b2, CHAIN_IGNORE_LENGTH);
}

// return false if overlap, 
bool nodesUnionable(node* node1, node* node2, int gapthreshold, int nodethreshold, int chain_ratio) {
  list<bzTag*> *tags1=node1->tags, *tags2=node2->tags;
  vector< vector<bzTag*>* > *chains1, *chains2;
  int curr1, curr2, size1, size2, mixed_chain=0, i;
  char *chained1, *chained2;
  
  chains1 = chainingTags(tags1, gapthreshold); 
  chains2 = chainingTags(tags2, gapthreshold);
  sort(chains1->begin(), chains1->end(), compar_chain_x2); // sort on Y
  sort(chains2->begin(), chains2->end(), compar_chain_x2);
  size1 = chains1->size();
  size2 = chains2->size();
  if ( size1 == 0 || size2 == 0 ) {
    delete_chains(chains1);
    delete_chains(chains2);
    return true;
  }
  chained1 = (char*)malloc(size1*sizeof(char));
  chained2 = (char*)malloc(size2*sizeof(char));
  for (i=0; i<size1; i++)
    chained1[i] = 'n';
  for (i=0; i<size2; i++)
    chained2[i] = 'n';
  
  long front_y1, front_y2, next_x1, next_x2;
  int ratio;
  for(curr1=curr2=0; curr1<size1 && curr2<size2; ) {
    if ( chain_overlap((*chains1)[curr1], (*chains2)[curr2]) == true ) { // overlap
      free(chained1);
      free(chained2);
      delete_chains(chains1);
      delete_chains(chains2);
      return false;
    }
    if ( chained1[curr1] == 'n' && chained2[curr2] == 'n' && (*((*chains1)[curr1]->begin()))->strand2 == (*((*chains2)[curr2]->begin()))->strand2 ) {
      if ( (*((*chains1)[curr1]->begin()))->strand2 == '+' ) {
	if ( (*((*chains1)[curr1]))[0]->b2 >  (*((*chains2)[curr2]))[0]->b2 ) {
	  front_y1 = (*((*chains2)[curr2]->end()-1))->e1;
	  front_y2 = (*((*chains2)[curr2]->end()-1))->e2;
	  next_x1  = (*((*chains1)[curr1]->begin()))->b1;
	  next_x2  = (*((*chains1)[curr1]->begin()))->b2;
	}
	else {
	  front_y1 = (*((*chains1)[curr1]->end()-1))->e1;
	  front_y2 = (*((*chains1)[curr1]->end()-1))->e2;
	  next_x1  = (*((*chains2)[curr2]->begin()))->b1;
	  next_x2  = (*((*chains2)[curr2]->begin()))->b2;
	  
	}
	if ( next_x1-front_y1>=0 && next_x1-front_y1<= nodethreshold && next_x2-front_y2>=0 && next_x2-front_y2<=nodethreshold ) {
	  if ( next_x1 - front_y1 < CHAIN_SMALL_GAP ) {
	    mixed_chain++;
	    chained1[curr1] = 'y';
	    chained2[curr2] = 'y';
	  }
	  else {
	    ratio = ( next_x2 - front_y2 + 1 )*100/( next_x1 - front_y1 );
	    if ( ratio <= 500 && ratio >= 20 ) {
	      mixed_chain++;
	      chained1[curr1] = 'y';
	      chained2[curr2] = 'y';
	    }
	  }
	} // if
      } // if
      else { // '-'
	if ( (*((*chains1)[curr1]))[0]->b2 >  (*((*chains2)[curr2]))[0]->b2 ) {
	  front_y1 = (*((*chains2)[curr2]->begin()))->b1;
	  front_y2 = (*((*chains2)[curr2]->begin()))->e2;
	  next_x1  = (*((*chains1)[curr1]->end()-1))->e1;
	  next_x2  = (*((*chains1)[curr1]->end()-1))->b2;
	}
	else {
	  front_y1 = (*((*chains1)[curr1]->begin()))->b1;
	  front_y2 = (*((*chains1)[curr1]->begin()))->e2;
	  next_x1  = (*((*chains2)[curr2]->end()-1))->e1;
	  next_x2  = (*((*chains2)[curr2]->end()-1))->b2;
	}
	if ( front_y1 - next_x1 >=0 && front_y1 - next_x1<= nodethreshold && next_x2-front_y2>=0 && next_x2-front_y2<=nodethreshold ) {
	  if ( front_y1 - next_x1 <= CHAIN_SMALL_GAP ) {
	    mixed_chain++;
	    chained1[curr1] = 'y';
	    chained2[curr2] = 'y';
	  }
	  else {
	    ratio = ( next_x2 - front_y2 + 1 )*100/( front_y1 - next_x1 );
	    if ( ratio <= 500 && ratio >= 20 ) {
	      mixed_chain++;
	      chained1[curr1] = 'y';
	      chained2[curr2] = 'y';
	    }
	  }
	}
      }// else
    } // if
   if ( (*((*chains1)[curr1]))[0]->b2 >  (*((*chains2)[curr2]))[0]->b2 )
      curr2++;      
    else
      curr1++;
  }// for
  free(chained1);
  free(chained2);
  delete_chains(chains1);
  delete_chains(chains2);

  if ( mixed_chain*100/size1 > chain_ratio && mixed_chain*100/size2 > chain_ratio)
    return true;
  return false;
}

void unionNodes(vector<node*>& nodes, int chain_ratio) {
  int size = nodes.size(), i;
  node* currNode, *nextNode;
  char* eraseArr;

  eraseArr = (char*)malloc(size*sizeof(char));
  memset(eraseArr, 'n', size);
  sort(nodes.begin(), nodes.end(), compar_node_start);
  for (i=0; i<size-1; i++) { // no need to process the last one
    currNode = nodes[i];
    nextNode = nodes[i+1];
    if ( nextNode->start - currNode->end > NODE_THRESHOLD )
      continue;
    // tags not overlap on y

    // diagonal
    if ( nodesUnionable(currNode, nextNode, GAP_THRESHOLD, 500, chain_ratio ) == true ) {
    //if ( nodesUnionable(currNode, nextNode, 3000, 3000, 10 ) == true ) {
    // union two nodes, update the tags, combine into the nextNode
      eraseArr[i] = 'y';
      nextNode->start = currNode->start;
      while ( currNode->tags->size() > 0 ){ 
	nextNode->tags->push_back(currNode->tags->back());
	currNode->tags->pop_back();
      }
    }
  }
  for (i=size-1; i>=0; i--)
    if ( eraseArr[i] == 'y' ) {
      clean_edges(nodes[i]->edges);
      delete nodes[i]->edges;
      delete nodes[i]->tags;
      delete nodes[i];
      nodes.erase( nodes.begin() + i );
    }
  free(eraseArr);
}

// need to construct edges of graph
void construct_graph(vector<node*>& nodes) {
  int i, nodesize = nodes.size();

  sort(nodes.begin(), nodes.end(), compar_node_start);

  for (i=0; i<nodesize; i++) {
    construct_graph_node(nodes, i);
  }// for each node
}


// assume e1 locates at left of e2 ( e1->b1 <= e2->b1)
long edges_overlap_X(edge* e1, edge* e2, int looser=0) {
  long a1, a2, b1, b2;

  a1 = (*(e1->tags->begin()))->b1;
  a2 = (*(e1->tags->end()-1))->e1;
  b1 = (*(e2->tags->begin()))->b1;
  b2 = (*(e2->tags->end()-1))->e1;
  if ( overlap(a1, a2, b1, b2, CHAIN_IGNORE_LENGTH, looser) == false ) {
    if ( a1 < b1 )
      return (a2/2+b1/2);
    else
      return (b2/2+a1/2);
  }
  /*
  over_beg = a1 > b1 ? a1 : b1;
  over_end = a2 < b2 ? a2 : b2;
  over1 = int(double((over_end-over_beg+1)/(a2-a1+1))*100);
  over2 = int(double((over_end-over_beg+1)/(b2-b1+1))*100);
  if ( over1 > 80 || over2 > 80 )
    return 0;
  if ( over1 < 20 || over2 < 20 )
    return (a2/2+b1/2);
  */
  return 0; 
}

/*
// assume tags in each edge are sorted by x1
// if ( edges of the same node connect to different nodes and they
// don't have much overlap, split
void splitLargeNode(vector<node*>& nodes, int currPos, vector<bzTag*>* tags) {
  node* n=nodes[currPos];
  vector<edge*>* edges = n->edges;
  int edgeNum = edges->size();
  int i, j, inner;
  long breakpoint = 0;

  inner=0;
  for (i=0; i<edgeNum; i++) {
    for (j=i+1; j<edgeNum; j++)
      if ( (breakpoint=edges_overlap_X( (*edges)[i], (*edges)[j])) > 0 ) {
	inner =1;
	break;
      }
    if ( inner == 1 )
      break;
  }
  if ( breakpoint == 0 )
    return;
  cutNode(nodes, currPos, breakpoint, tags);
  splitLargeNode(nodes, currPos, tags);
}

void splitLargeNodes(vector<node*>& nodes, vector<bzTag*>* tags) {
  int i, size=nodes.size();

  for (i=0; i<size; i++)
    splitLargeNode(nodes, i, tags);
}
*/

/*
vector< vector<int>* >* get_ccs(vector<node*>& nodes) {
  int size = nodes.size(), i, j, k, count;
  char* visitedArr = (char*)malloc(size*sizeof(char));
  vector< vector<int>* >* ccs;
  vector< vector<int>* >::iterator ccIt;
  vector<int>* cc;
  vector<int>::iterator intIt;
  list<edge*>::iterator edgeIt;
  list<edge*>* edges;
  int edgeSize;
  node* currNode;
  int ccCount;
  int* node2ccMap;
  vector<int> fuseCCs;

  node2ccMap = (int*)malloc(size*sizeof(int));
  for (i=0; i<size; i++)
    node2ccMap[i] = -1;

  memset(visitedArr, 'n', size);

  ccCount = 0;
  ccs = new vector< vector<int>* >;
  for (i=0; i<size; i++) {
    if (visitedArr[i] == 'y' )
      continue;
    fuseCCs.clear();
    cc = new vector<int>;
    ccCount++;
    cc->push_back(i);
    node2ccMap[i]=ccCount-1; // record the index of cc of the curr node
    ccs->push_back(cc);
    visitedArr[i] = 'g';     // grey
    count = 1;
    for (j=0; j<count; j++) {
      currNode = nodes[(*cc)[j]];
      edges = currNode->edges;
      edgeSize = edges->size();      
      for (edgeIt=edges->begin(); edgeIt != edges->end(); ++edgeIt) {
	if ( visitedArr[(*edgeIt)->node_index] == 'y' && node2ccMap[(*edgeIt)->node_index] != ccCount-1) {
	  for (intIt = fuseCCs.begin(); intIt < fuseCCs.end(); intIt++)
	    if ( *intIt == node2ccMap[(*edgeIt)->node_index] )
	      break;
	  if ( intIt == fuseCCs.end() )
	    fuseCCs.push_back( node2ccMap[(*edgeIt)->node_index] );
	}
	else if ( visitedArr[(*edgeIt)->node_index] == 'n' ) {
	  cc->push_back((*edgeIt)->node_index);
	  visitedArr[(*edgeIt)->node_index] = 'g';
	  count++;
	}
      } // for k
      visitedArr[ (*cc)[j] ] = 'y';
      node2ccMap[ (*cc)[j] ] = ccCount-1;
    } // for j
    

    ///---- no use
    if ( fuseCCs.empty() )
      continue;
    fuseCCs.push_back(ccCount-1);
    sort(fuseCCs.begin(), fuseCCs.end(), compar_int);
    // cat everything to the first cc in fuseNodes
    int fuseSize=fuseCCs.size(), ccSize;
    vector<int>* currCC;
    for (k=fuseSize-1; k>0; k--) {
      currCC = (*ccs)[fuseCCs[k]];
      ccSize = currCC->size();
      for (j=ccSize-1; j>=0; j--) {
	(*ccs)[fuseCCs[0]]->push_back( (*currCC)[j] );
	currCC->pop_back();
      } // for j
    } // for k
    for (k=fuseSize-1; k>0; k--) {
      delete (*ccs)[ fuseCCs[k] ];
      ccs->erase( ccs->begin() + fuseCCs[k] );
      ccCount--;
    }
    // node2cc is changed
    for (k=1; k<fuseSize; k++)
      for (j=0; j<size; j++)
	if ( node2ccMap[j] == fuseCCs[k] )
	  node2ccMap[j] = fuseCCs[0];
    /// ------end of no use
  }// for i
  free(node2ccMap);
  free(visitedArr);
  int ccsSize=ccs->size();
  for (j=0; j<ccsSize; j++) {
    cc = (*ccs)[j];
    sort(cc->begin(), cc->end(), compar_int);
  }
  return ccs;
}
*/

// using breadth first search to get ccs
vector< vector<int>* >* get_ccs(vector<node*>& nodes) { 
  vector< vector<int>* >* ccs = new vector< vector<int>* >;
  vector<int>* cc;
  deque<int> walkqueue;
  deque<int> pathqueue;
  list<edge*>* edges;
  list<edge*>::iterator edgeIt;
  int nodesize = nodes.size(), i, k, curr;

  for (k=0; k<nodesize; k++) {
    nodes[k]->visited = 'n';
    nodes[k]->parent = -1;
  }
  

  for (i=0; i<nodesize; i++) {
    if ( nodes[i]->visited == 'y' )
      continue;
    cc = new vector<int>;
    cc->push_back(i);
    walkqueue.push_back(i);
    nodes[i]->visited  = 'y';
    while ( !walkqueue.empty() ) {
      curr = walkqueue.front();
      walkqueue.pop_front();
      edges = nodes[curr]->edges;
      for (edgeIt = edges->begin(); edgeIt != edges->end(); ++edgeIt)
	if ( nodes[ (*edgeIt)->node_index ]->visited == 'n' ) {
	  pathqueue.push_front( (*edgeIt)->node_index );
	  k = curr;
	  while ( k != -1 ) {
	    pathqueue.push_front( k );
	    k = nodes[k]->parent;
	  }
	  if ( consistent(nodes, pathqueue, 1) == -1 ) {
	    walkqueue.push_back( (*edgeIt)->node_index );
	    cc->push_back( (*edgeIt)->node_index );
	    nodes[(*edgeIt)->node_index]->visited = 'y';
	    nodes[(*edgeIt)->node_index]->parent = curr;
	  }
	  pathqueue.clear();
	}
    }
    ccs->push_back(cc);
  }
  int ccsSize=ccs->size();
  for (i=0; i<ccsSize; i++) {
    cc = (*ccs)[i];
    sort(cc->begin(), cc->end(), compar_int);
  }
  return ccs;
}

void display_ccs(vector< vector<int>* >* ccs, vector<node*>& nodes, ostream& os) {  
  // print out
  int k, size, i, j;
  vector<int>::iterator intIt;

  size = ccs->size();
  for (k=0; k<size; k++)
    sort( (*ccs)[k]->begin(), (*ccs)[k]->end(), compar_int);
  for (i=j=0; i<size; i++) {
    if ( (*ccs)[i]->size() < 2 )
      continue;
    os << j++ << ": " << (*ccs)[i]->size() << ": ";
    for (intIt = (*ccs)[i]->begin(); intIt < (*ccs)[i]->end(); intIt++)
      os << (*intIt) << ' ';
    os << endl;
  }
  return;
}

void make_bidirectional_graph(vector<node*>& nodes) {
  int nodeSize=nodes.size(), edgeSize1, edgeSize2, j, k, i;
  list<edge*>* edges1, *edges2;
  list<edge*>::iterator edgeIt1, edgeIt2, edgeIt1next;
  list<edge*>::reverse_iterator rEdgeIt1, rEdgeIt1next;
  char *eraseArr;
  int index2;

  eraseArr = (char*)malloc(nodeSize*sizeof(char)); // to save malloc/free
  for (j=0; j<nodeSize; j++) {
    edges1 = nodes[j]->edges;
    edgeSize1 = edges1->size();
    memset(eraseArr, 'n', edgeSize1);
    for (k=0, edgeIt1=edges1->begin(); k<edgeSize1; ++edgeIt1, k++) { // for each edge of node1
      index2 = (*edgeIt1)->node_index;
      if ( index2 == j ) // itself
	continue;
      edges2 = nodes[index2]->edges;
      edgeSize2 = edges2->size();
      for (i=0, edgeIt2=edges2->begin(); i<edgeSize2; ++edgeIt2, i++) 
	if ( (*edgeIt2)->node_index == j )
	  break;
      if ( i == edgeSize2 )  // not found, need to erase
	eraseArr[k] = 'y';
    }
    edgeIt1 = edges1->end();
    edgeIt1--;
    for (k=edgeSize1-1; k>=0; k--, edgeIt1=edgeIt1next) {
      edgeIt1next = edgeIt1;
      edgeIt1next--;
      if ( eraseArr[k] == 'y' ) {
	(*edgeIt1)->tags->clear();
	delete (*edgeIt1)->tags;
	delete *edgeIt1;
	edges1->erase( edgeIt1 );
      }
    }
  }
  free(eraseArr);
}

// Once this function is called in main, bug not solved.
void remove_self_nodes(vector<node*>& nodes) {
  int nodeSize=nodes.size(), edgeSize, j;
  list<edge*>* edges;
  vector<bzTag*>* tags;
  char *eraseArr;
  
  eraseArr = (char*)malloc(nodeSize*sizeof(char));
  memset(eraseArr, 'n', nodeSize);
  for (j=0; j<nodeSize; j++) {
    edges = nodes[j]->edges;
    edgeSize = edges->size();
    if ( edgeSize == 1 ) {
      tags = (*(edges->begin()))->tags;
      if ( tags->size() == 0 )
	;
      //cerr << "empty edge: " << j << ' ' << (*edges)[0]->node_index << endl;
      else if ( (*(tags->begin()))->b1 == (*(tags->begin()))->b2 ) {
	eraseArr[j] = 'y';
      }
    }
  }
  for (j=nodeSize-1; j>=0; j--) {
    if ( eraseArr[j] == 'y' ) {
      (*(nodes[j]->edges->begin()))->tags->clear();
      clean_edges( nodes[j]->edges );
      delete nodes[j]->edges;
      nodes[j]->tags->clear();
      delete nodes[j]->tags;
      delete *(nodes.begin() + j);
      nodes.erase( nodes.begin() + j );
    }
  }
  free(eraseArr);
}

  // add the identical self alignment tags
void addSelfAlignment(vector<node*>& nodes) {
  int k, size=nodes.size();
  bzTag* newtag;

  for (k=0; k<size; k++) {
    newtag = new bzTag;
    newtag->b1 = newtag->b2 = nodes[k]->start;
    newtag->e1 = newtag->e2 = nodes[k]->end;
    newtag->pci = 100;
    newtag->strand2 = '+';
    //tags->push_back(newtag);
    nodes[k]->tags->push_back(newtag);
  }
}

void rmSelfAlignment(vector<node*>& nodes) {
  list<bzTag*>* tags;
  list<bzTag*>::iterator tagIt, tagItnext;
  list<bzTag*>::reverse_iterator rTagIt, rTagItnext;
  int i, k, nodesize=nodes.size(), tagsize;
  char* eraseArrTag;

  for (k=0; k<nodesize; k++) {
    tags = nodes[k]->tags;
    tagsize = tags->size();
    eraseArrTag = (char*)malloc(tagsize*sizeof(char));
    memset(eraseArrTag, 'n', tagsize);
    for (i=0, tagIt=tags->begin(); i<tagsize; ++tagIt, i++)
      if ( (*tagIt)->b1 == (*tagIt)->b2 && (*tagIt)->strand2 == '+' )
	eraseArrTag[i] = 'y';
    tagIt = tags->end();
    tagIt--;
    for (i=tagsize-1; i>=0; i--, tagIt=tagItnext) {
      tagItnext = tagIt;
      tagItnext--;
      if ( eraseArrTag[i] == 'y' ){
	delete *tagIt;
	tags->erase( tagIt );
      }
    }
    free(eraseArrTag);
  }
}

void cpTags(node* source, node* target) {
  int tagSize, k;
  list<bzTag*>* tags = source->tags;
  list<bzTag*>::iterator tagIt;

  tagSize = tags->size();
  for (tagIt=tags->begin(), k=0; k<tagSize; ++tagIt, k++)
    target->tags->push_back(*tagIt);
}

void cpEdges(node* source, node* target) {
  int edgeSize, k;
  list<edge*>* edges = source->edges;
  list<edge*>::iterator edgeIt;
  
  edgeSize = edges->size();
  for (edgeIt=edges->begin(), k=0; k<edgeSize; ++edgeIt, k++)
    target->edges->push_back( *edgeIt);
}

void fuseNodes(vector<node*>& nodes, vector<int>* cc, char* eraseArr) {

  int ccSize, k;

  sort(cc->begin(), cc->end(), compar_int);
  ccSize = cc->size();

  for (k=1; k<ccSize; k++) {
    if ( (*cc)[k]==(*cc)[k-1]+1 && nodesUnionable( nodes[ (*cc)[k-1] ], nodes[ (*cc)[k] ], 5000, 5000, 20 ) ) {
      nodes[(*cc)[k]]->start = nodes[(*cc)[k-1]]->start;
      cpTags(nodes[(*cc)[k-1]], nodes[(*cc)[k]]);
      eraseArr[(*cc)[k-1]] = 'y';
      nodes[(*cc)[k-1]]->tags->clear();
    }
  }
}
// make sure cc are sorted.
void finalizeNodes(vector<node*>& nodes, vector< vector<int>* >* ccs) {
  char *eraseArr;
  int ccSize=ccs->size(), k, nodeSize=nodes.size();

  eraseArr = (char*)malloc(nodeSize*sizeof(char));
  for (k=0; k<nodeSize; k++)
    eraseArr[k] = 'n';

  for (k=0; k<ccSize; k++)
    fuseNodes(nodes, (*ccs)[k], eraseArr);

    // clear nodes marked
  for (k=nodeSize-1; k>=0; k--)
    if ( eraseArr[k] == 'y' ) {
      clean_edges(nodes[k]->edges);
      nodes[k]->tags->clear();
      delete nodes[k]->tags;
      delete nodes[k]->edges;
      delete nodes[k];
      nodes.erase( nodes.begin() + k );
    }
  free(eraseArr);
}

bool compar_cc(vector<int>* cc1, vector<int>* cc2) {
  return cc1->size() > cc2->size();
}

/* original at 103106
long determine_breakpoint(node* n, int looser=0) {
  int i, edgesize;
  long breakpoint;
  list< edge* >* edges;
  list< edge* >::iterator edgeIt, edgeIt2, endIt;
  vector<bzTag*>* tags;

  if ( !n->edges->empty() ) {
    edges = n->edges;
    edgesize = edges->size();
    for (edgeIt=edges->begin(), i=0; i<edgesize; ++edgeIt, i++) {
      tags = (*edgeIt)->tags;
      sort(tags->begin(), tags->end(), compar_tag_x1);
    }
    edges->sort(compar_edge_x1);
    if ( looser == 0 ) { // only compare the first edge with rest
      endIt = edges->begin();
      endIt++;
    }
    else { // all to rest

      endIt = edges->end();
    }
    for (edgeIt=edges->begin(); edgeIt != endIt; ++edgeIt) {
      edgeIt2 = edgeIt;
      edgeIt2++;
      for (; edgeIt2 != edges->end(); ++edgeIt2)
	if ( (breakpoint = edges_overlap_X( *edgeIt, *edgeIt2, looser )) > 0 )
	  break;
      if ( edgeIt2 != edges->end() ) {
	if ( breakpoint < n->start || breakpoint > n->end )
	  printf("hello"); 
	// do minimum error optimized breakpoint.
	return breakpoint;
      }
    }
  }
  return 0;
}
*/

long determine_breakpoint(node* n, int looser=0) {
  int i, edgesize;
  long breakpoint;
  list< edge* >* edges;
  list< edge* >::iterator edgeIt, edgeIt2, endIt;
  vector<bzTag*>* tags;
  long start_range, end_range, edge_beg, edge_end;
  vector<long> ends, starts;

  if ( !n->edges->empty() ) {
    edges = n->edges;
    edgesize = edges->size();
    for (edgeIt=edges->begin(), i=0; i<edgesize; ++edgeIt, i++) {
      tags = (*edgeIt)->tags;
      sort(tags->begin(), tags->end(), compar_tag_x1);
    }
    edges->sort(compar_edge_x1);
    if ( looser == 0 ) { // only compare the first edge with rest
      endIt = edges->begin();
      endIt++;
    }
    else { // all to rest

      endIt = edges->end();
    }
    start_range = end_range = -1;
    for (edgeIt=edges->begin(); edgeIt != endIt; ++edgeIt) {
      edgeIt2 = edgeIt;
      edgeIt2++;
      for (; edgeIt2 != edges->end(); ++edgeIt2)
	if ( (breakpoint = edges_overlap_X( *edgeIt, *edgeIt2, looser )) > 0 ) {
	  edge_beg = (*edgeIt)->tags->front()->b1;
	  edge_end = (*edgeIt)->tags->back()->e1;
	  start_range = edge_beg/2 + edge_end/2;
	  edge_beg = (*edgeIt2)->tags->front()->b1;
	  edge_end = (*edgeIt2)->tags->back()->e1;
	  end_range = edge_beg/2 + edge_end/2;
	  if ( start_range < end_range )
	    break;
	}
      if ( edgeIt2 != edges->end() ) {
	//if ( breakpoint < n->start || breakpoint > n->end )
	  //printf("hello"); 
	// do minimum error optimized breakpoint.
	//return breakpoint;
	break;	
      }
    }
    // 
    if ( start_range >= 0 ) {
      //ends = new vector<long>;
      //starts = new vector<long>;
      for (edgeIt=edges->begin(); edgeIt != endIt; ++edgeIt) { 
	edge_beg = (*edgeIt)->tags->front()->b1;
	edge_end = (*edgeIt)->tags->back()->e1;
	starts.push_back(edge_beg);
	ends.push_back(edge_end);
	//if ( (edge_beg < start_range) && (edge_end >= start_range) && (edge_end < end_range) ) 
	//starts.push_back(edge_end);
	//if ( (edge_beg > start_range) && (edge_beg <= end_range) && (edge_end > end_range) )
	//ends.push_back(edge_beg);
      }
      breakpoint = graph_cutpoint(starts, ends, start_range, end_range);
      //cerr << "breakpoing: " << breakpoint << endl;
      return breakpoint;
    }
  }
  return 0;
}

int adjustLargeNodes(vector<node*>& nodes) {
  int i, nodesize=nodes.size(), adjust=0;
  long breakpoint;
 
  for (i=0; i<nodesize; i++) {
    if ( nodes[i]->end - nodes[i]->start + 1 < MIN_CHAIN )
      continue;
    if ( (breakpoint=determine_breakpoint(nodes[i])) > 0 ) {
      adjust=1;
      if ( breakpoint > nodes[i]->end || breakpoint < nodes[i]->start )
	printf("hello");
      cutNode(nodes, i, breakpoint, 0);
    }
  }
  sort(nodes.begin(), nodes.end(), compar_node_start);
  return adjust;
}

void checkTags(vector<node*>& nodes) {
  int i, nodesize=nodes.size(), tagsize, j;
  list<bzTag*>* tags;
  list<bzTag*>::iterator tagIt;

  for (i=0; i<nodesize; i++) {
    if ( nodes[i] == NULL )
      continue;
    tags = nodes[i]->tags;
    tagsize = tags->size();
    for (tagIt = tags->begin(), j=0; tagIt != tags->end(); ++tagIt, ++j) {
      if ( (*tagIt)->b1 > (*tagIt)->e1 )
	cerr << "Wrong tag1: " << i << ' ' << j << endl;
      if ( (*tagIt)->b2 > (*tagIt)->e2)
	cerr << "Wrogn tag2: " << i << ' ' << j << endl;
    }
  }
}

bool compar_int_ptr(int* i1, int* i2) {
  return *i1 < *i2;
}

bool compar_vec_int( vector<int>* v1, vector<int>* v2 ) {
  return (*v1)[0] < (*v2)[0];
}

bool testCommonNode( list<edge*>& edges1, list<edge*>& edges2) {
  list<edge*>::iterator edgeIt1, edgeIt2;
  
  for (edgeIt1=edges1.begin(); edgeIt1 != edges1.end(); ++edgeIt1)
    for (edgeIt2=edges2.begin(); edgeIt2 != edges2.end(); ++edgeIt2)
      if ( (*edgeIt1)->node_index == (*edgeIt2)->node_index )
	return true;
  return false;

}

void sortCCS(vector< vector<int>* >* ccs) {
  vector<int*> vec;
  int* ptr;
  int size=ccs->size(), i, intSize, j;

  for (i=0; i<size; i++) {
    intSize = (*ccs)[i]->size();
    for (j=0; j<intSize; j++) {
      ptr = (int*)malloc(sizeof(int));
      *ptr = (*(*ccs)[i])[j];
      vec.push_back(ptr);
    }
    sort(vec.begin(), vec.end(), compar_int_ptr);
    (*ccs)[i]->clear();
    for (j=0; j<intSize; j++)
      (*ccs)[i]->push_back( *(vec[j]) );

    intSize = vec.size();
    for (j=0; j<intSize; j++)
      free(vec[j]);
    vec.clear();
  }
  sort(ccs->begin(), ccs->end(), compar_vec_int);
}

void collapse(vector< vector<int>* >& ccs, vector<node*>& nodes, int collaps_gap) {
  int  tagSize, ccsSize, nodeSize, i, j, currSize, nextSize, currIndex;
  list<bzTag*>::iterator tagIt;
  list<bzTag*>* tags;

  ccsSize = ccs.size();
  for (i=0; i<ccsSize-1; i++) {
    currSize = ccs[i]->size();
    nextSize = ccs[i+1]->size();
    if ( currSize != nextSize )
      continue;
    for (j=0; j<currSize; j++)
      if ( (*(ccs[i]))[j] + 1 != (*(ccs[i+1]))[j] )
	break;
    if ( j < currSize )
      continue;
    for (j=0; j<currSize; j++) {
      currIndex = (*(ccs[i]))[j];
      if ( nodes[currIndex+1]->start - nodes[currIndex]->start > collaps_gap )
	break;
    }
    if ( j < currSize )
      continue;
    
    for (j=0; j<currSize; j++) {
      currIndex = (*(ccs[i]))[j];
      if ( testCommonNode( *(nodes[currIndex]->edges), *(nodes[currIndex+1]->edges)) )
	break;
    }
    if ( j < currSize )
      continue;
    
    for (j=0; j<currSize; j++) {
      currIndex = (*(ccs[i]))[j];
      nodes[currIndex+1]->start = nodes[currIndex]->start;
      tagSize = nodes[currIndex]->tags->size();
      tags = nodes[currIndex]->tags;
      for (tagIt=tags->begin(); tagIt != tags->end(); ++tagIt)
	nodes[currIndex+1]->tags->push_back( *tagIt );
      nodes[currIndex]->tags->clear();
      delete nodes[currIndex]->tags;
      clean_edges(nodes[currIndex]->edges);
      delete nodes[currIndex]->edges;
      delete nodes[currIndex];
      nodes[currIndex] = NULL;
    }
  }
  nodeSize = nodes.size();
  for (i=nodeSize-1; i>=0; i--)
    if ( nodes[i] == NULL )
      nodes.erase( nodes.begin() + i );
}

// all edges become empty, but edges objects are not deleted
void clean_all_edges(vector<node*>& nodes) {
  int i, size=nodes.size();

  for (i=0; i<size; i++)
    if ( nodes[i] != NULL )
      clean_edges(nodes[i]->edges);
}

vector<bzTag*>* transfer_all_tags(vector<node*>& nodes) {
  int i, j, nodeSize=nodes.size(), tagsize;
  vector<bzTag*>* tags = new vector<bzTag*>;
  
  for (i=0; i<nodeSize; i++) {
    if ( nodes[i] != NULL ) {
      tagsize = nodes[i]->tags->size();
      for (j=0; j<tagsize; j++) {
	tags->push_back( nodes[i]->tags->back() );
	nodes[i]->tags->pop_back();
      }
    }
  }
  return tags;
}

long coverageLength(vector<node*>& nodes) {
  long len=0;
  int i, size=nodes.size();

  for (i=0; i<size; i++)
    len += (nodes[i]->end - nodes[i]->start + 1 );

  return len;
}

// test whether there's an edge between node i and j
bool edgeOrNot(vector<node*>& nodes, int i, int j) {
  list<edge*>* edges=nodes[i]->edges;
  list<edge*>::iterator edgeIt;
  
  for (edgeIt = edges->begin(); edgeIt != edges->end(); ++edgeIt )
    if ( (*edgeIt)->node_index == j )
      return true;
  return false;
}

int commonNodeOrNot(vector<node*>& nodes, int i, int j) {
  int smaller, larger, edge1size, edge2size, common=0;
  list<edge*>* edges1, *edges2;
  list<edge*>::iterator edgeIt1, edgeIt2;

  if ( nodes[i]->edges->size() > nodes[j]->edges->size() ) {
    smaller = j;
    larger = i;
  }
  else {
    smaller = i;
    larger = j;
  }
  edges1 = nodes[smaller]->edges;
  edges2 = nodes[larger]->edges;
  edge1size = edges1->size();
  edge2size = edges2->size();
  
  for (edgeIt1=edges1->begin(), edgeIt2=edges2->begin(); edgeIt1 != edges1->end() && edgeIt2 != edges2->end(); ) {
    while ( edgeIt2 != edges2->end() && (*edgeIt2)->node_index < (*edgeIt1)->node_index-1 )
      ++edgeIt2;
    if ( edgeIt2 == edges2->end() )
      break;
    if ( ((*edgeIt2)->node_index == (*edgeIt1)->node_index-1 )
	 || ((*edgeIt2)->node_index == (*edgeIt1)->node_index )
	 || ((*edgeIt2)->node_index == (*edgeIt1)->node_index+1) ) {
      if ((*edgeIt2)->tags->front()->strand2 == (*edgeIt1)->tags->front()->strand2)
	common++;     
      edgeIt1++;
    }
    while ( edgeIt1 != edges1->end() && (*edgeIt1)->node_index < (*edgeIt2)->node_index-1 )
      ++edgeIt1;
    if ( edgeIt1 == edges1->end() )
      break;
    if ( ((*edgeIt1)->node_index == (*edgeIt2)->node_index-1 )
	 || ((*edgeIt1)->node_index == (*edgeIt2)->node_index )
	 || ((*edgeIt1)->node_index == (*edgeIt2)->node_index+1) ) {
      if ((*edgeIt2)->tags->front()->strand2 == (*edgeIt1)->tags->front()->strand2)     
	common++;
      edgeIt1++;
    }
  }
  if ( common > 0 && (common*100/edge1size >= 20 || common*100/edge2size >= 20 ) )
    return 1;
  else
    return 0;
}
  
void sort_edge_x2(vector<node*>& nodes) {
  int size=nodes.size(), i;

  for (i=0; i<size; i++) 
    nodes[i]->edges->sort(compar_edge_x2);
}

void removeTagFromEdge(node* n, edge* e);

void cut_graph(vector<node*>& nodes, vector<cut_edge*>* cut_edges) {
  int cutsize=cut_edges->size(), edgesize, i, node1, node2;
  list<edge*>* edges;
  list<edge*>::iterator edgeIt;

  for (i=0; i<cutsize; i++){ 
    node1 = (*cut_edges)[i]->node1;
    node2 = (*cut_edges)[i]->node2;
    edges = nodes[node1]->edges;
    edgesize = edges->size();
    for (edgeIt=edges->begin(); edgeIt != edges->end(); ++edgeIt)
      if ( (*edgeIt)->node_index == node2 )
	break;
    if ( edgeIt != edges->end() ) {
      removeTagFromEdge(nodes[node1], *edgeIt);
      delete *edgeIt;
      edges->erase( edgeIt );
    }
    else
      cerr << "Error!\n";
  }
}

// suppose markArr has been initialized with 'u's
// suppose each cc is sorted
int error_measurement(vector<node*>& nodes, vector< vector<int>* >* ccs, char* markArr) {
  vector<int>* cc;

  int total_cc = ccs->size();
  int ccSize, i, error=0, error0, total_error0=0, total_error=0, j, error1, error2, total_error1=0, total_error2=0, common, consis;

  for (i=0; i<total_cc; i++) {
    cc = (*ccs)[i];
    ccSize = cc->size();
    error=error1 = error2 = error0 =0;
    for (j=0; j<ccSize-1; j++) {
      if ( (*cc)[j]+1 == (*cc)[j+1] ) { // adjacent
	if ( nodes[(*cc)[j+1]]->start - nodes[(*cc)[j]]->end < NODE_THRESHOLD) {
	  if ( !edgeOrNot(nodes, (*cc)[j], (*cc)[j+1]) ) { // no direct edge 
	    consis = breadth_first_shortest_path(nodes, (*cc)[j], (*cc)[j]+1);
	    if ( consis == 2 ) {
	      
	      error++;
	      cerr << ' ' << (*cc)[j];
	      common = commonNodeOrNot(nodes, (*cc)[j], (*cc)[j+1]);
	      if ( common == 0 ) { // no common neighbors or below a threshold
		error0++;
		cerr << "(error0: " << (*cc)[j] << ": ";
		
		cerr << ") ";
		markArr[ (*cc)[j] ] = 's';
	      }
	      else {//if ( common == 1 ) { // sharing many neighors
		error1++;
		markArr[ (*cc)[j] ] = 'm';
	      }
	    }
	  }
	}
      }
    }
    cerr << endl << i << " error: " << error << " error1: " << error1 << " error2: " << error2 << " error0: " << error0 <<  endl;
    total_error += error;
    total_error0 += error0;
    total_error1 += error1;
    total_error2 += error2;
  }
  cerr << "total error: " << total_error << " total error1: " << total_error1 << " total_error2: " << total_error2  << " total_error0: " << total_error0 << endl;
  return error;
}

void removeTagFromEdge(node* pNode, edge* pEdge) {
  vector<bzTag*>* tags1;
  list<bzTag*>* tags2;
  vector<bzTag*>::iterator tagIt1;
  list<bzTag*>::iterator tagIt2, tagIt2next;
  list<bzTag*>::reverse_iterator rTagIt2, rTagIt2next;
  int tagsize1, tagsize2, j, k;
  char* eraseArr;

  //erase tags in (*edges1)[m]
  tags1 = pEdge->tags;
  tags2 = pNode->tags;
  tagsize1 = tags1->size();
  tagsize2 = tags2->size();
  eraseArr = (char*)malloc(tagsize2*sizeof(char));
  memset(eraseArr, 'n', tagsize2);
  for (tagIt1=tags1->begin(); tagIt1 != tags1->end(); ++tagIt1)
    for (k=0, tagIt2=tags2->begin(); tagIt2 != tags2->end(); ++tagIt2, ++k)
      if ( (*tagIt1)->b1 == (*tagIt2)->b1 && (*tagIt1)->b2 == (*tagIt2)->b2 )
	eraseArr[k] = 'y';

  pEdge->tags->clear();

  tagIt2 = tags2->end();
  tagIt2--;
  for (j=tagsize2-1; j>=0; j--, tagIt2 = tagIt2next) {
    tagIt2next = tagIt2;
    tagIt2next--;
    if ( eraseArr[j] == 'y' ) {
      delete (*tagIt2);
      tags2->erase( tagIt2 );
    }
  }
  free(eraseArr);
}

// split first then merge, otherwise, the merging maybe splitted again in the same cycle
void mergeGraph(vector<node*>& nodes, char* markArr, vector< vector<int>* >* ccs) {
  int nodeSize = nodes.size(), i, tagsize, j;

  for (i=0; i<nodeSize-1; i++) {
    if ( markArr[i] == 'm' ) {
      if ( nodes[i] == NULL || nodes[i+1] == NULL )
	continue;
      nodes[i+1]->start = nodes[i]->start;

      tagsize = nodes[i]->tags->size();
      for (j=0; j<tagsize; j++) {
	nodes[i+1]->tags->push_back( nodes[i]->tags->back() );
	nodes[i]->tags->pop_back();
      }
      clean_edges(nodes[i]->edges);
      delete nodes[i]->edges;
      delete nodes[i]->tags;      
      delete nodes[i];
      nodes[i] = NULL;
    }
  }
  for (i=nodeSize-1; i>=0; i--)
    if ( nodes[i] == NULL )
      nodes.erase( nodes.begin() + i );
}

void free_cut_edges(vector<cut_edge*>* cut_edges) {
  int size=cut_edges->size(), i;

  for (i=0; i<size; i++)
    delete( (*cut_edges)[i] );
  cut_edges->clear();
  delete cut_edges;
}

void splitGraph(vector<node*>& nodes, char* markArr, vector< vector<int>* >* ccs) {
  int ccsSize=ccs->size(), ccsize, j, k, node_index;
  vector<int>* cc;
  vector<cut_edge*>* cut_edges;
  int pathOrNot, counter, large_cc_size;

  if ( ccs->size() == 0 )
    return;
  large_cc_size = (*ccs)[0]->size();
  initialize_flow_graph(nodes);
  for (k=0; k<ccsSize; k++) { // for each cc, Gc is the same
    counter = 0;
    cc = (*ccs)[k];
    ccsize = cc->size();
    for (j=0; j<ccsize; j++) {      
      node_index = (*cc)[j];
      if ( markArr[ node_index ] == 's' ) {
	if ( nodes[node_index] == NULL || nodes[node_index+1] == NULL )
	  continue;
	pathOrNot = breadth_first_shortest_path(nodes, node_index, node_index+1);
	if ( pathOrNot == 2) {
	  cut_edges = max_flow_min_cut(nodes, cc, (*cc)[j], (*cc)[j]+1);
	  cut_graph(nodes, cut_edges);
	  free_cut_edges(cut_edges);
	  counter++;
	}
      }
    }
  }
  delete_flow_graph();
}

void delete_ccs(vector< vector<int>* >* ccs) {
  int ccsSize, i;
  ccsSize = ccs->size();
  for (i=0; i<ccsSize; i++) {
    (*ccs)[i]->clear();
    delete (*ccs)[i];
  }
  ccs->clear();
  delete ccs;
}


void make_good_graph(vector<node*>& nodes) {
  vector<bzTag*>* tags;

  clean_all_edges(nodes);
  tags = transfer_all_tags(nodes);
  groupTags2(nodes, tags); // tags shall be empty now
  maintenance_nodes(nodes);
  construct_graph(nodes);
  make_bidirectional_graph(nodes);

  tags->clear();
  delete tags;
}
  
int mergeSplitIteration(vector<node*>& nodes, void (*mergeSplit)(vector<node*>& nodes, char* markArr, vector< vector<int>* >* ccs)) {
  vector< vector<int>* >* ccs;
  int nodeSize = nodes.size(), i, ccsSize;
  char* markArr = (char*)malloc(nodeSize*sizeof(char));

  make_good_graph(nodes);
  ccs = get_ccs(nodes); 
  sort(ccs->begin(), ccs->end(), compar_cc); 
  ccsSize = ccs->size();
  cerr << "nodes size: " << nodes.size() << " ccs size: " << ccsSize << " first cc size: " << (*ccs)[0]->size() << endl;
  for (i=0; i<nodeSize; i++)
    markArr[i] = 'u';

  sort_edge_x2(nodes);
  error_measurement(nodes, ccs, markArr);
  mergeSplit(nodes, markArr, ccs);

  free(markArr);
  delete_ccs(ccs);
  return ccsSize;
}

void clear_nodes(vector<node*>& nodes) {
  int i, nodesize=nodes.size();
  list<bzTag*>* tags;
  list<bzTag*>::iterator tagIt, tagItnext;

  clean_all_edges(nodes);
  for (i=0; i<nodesize; i++) {
    tags = nodes[i]->tags;
    for (tagIt=tags->begin(); tagIt != tags->end(); tagIt=tagItnext) {
      tagItnext = tagIt;
      tagItnext--;
      delete *tagIt;
    }
    tags->clear();
    delete tags;
    delete nodes[i]->edges;
    // remove tags;
    delete nodes[i];
  }
  nodes.clear();
}


void program_display(vector<node*>& nodes, const char* fnstring) {
  char filename[1000];
  vector< vector<int>* >* ccs;
  vector<int>* cc;
  ofstream of;
  long coverage;
  int i, j, ccSize, ccsSize, nodeSize, consistency_count=0, consistency_total=0;

  ccs = get_ccs(nodes);
  sort(ccs->begin(), ccs->end(), compar_cc);
  
  coverage = coverageLength(nodes);
  cerr << " node size after last union: " << nodes.size() << ' ' << coverage << " cc size: " << ccs->size() << endl;
  
  sprintf(filename, "%s.nodes", fnstring);
  of.open(filename);  
  printNodes(nodes, of);
  of.close();

  sprintf(filename, "%s.edges", fnstring);
  of.open(filename);
  printNodeDetails(nodes, of);
  of.close();
  
  sprintf(filename, "%s.cccs", fnstring);
  of.open(filename);
  display_ccs(ccs, nodes, of);
  of.close();

  ccsSize = ccs->size();
  nodeSize = nodes.size();
  allocate_ccArr(nodeSize);
  for (i=0; i<ccsSize; i++) {
    consistency_total = 0;
    cc = (*ccs)[i];
    ccSize = cc->size();
    initialize_ccArr(nodeSize, cc);
    if ( ccSize > 1 ) {
      for (j=0; j<ccSize; j++) {
	consistency_count = breadth_first_walk_tree(nodes, (*cc)[j], cc);
	consistency_total += consistency_count;
      }
      cerr << "#" << i << ' ' << ccSize << ", Consistency " << consistency_total/2 << " out of " << ccSize*(ccSize-1)/2 << " ratio: " << consistency_total*100/(ccSize*(ccSize-1)) << endl;
    }
  }
  free_ccArr();
  delete_ccs(ccs);
}

int main (int argc, char** argv) {
  char filename[500];
  vector<bzTag*> *tags, *alltags;
  vector<node*> nodes;
  vector< vector<int>* >* ccs;
  ofstream ofdup, ofrep, of;
  int self=1, iter_count, iter=REFINE_ITERATION;
  int oldsize=0, newsize;
  long coverage;

  if ( argc < 2 ) {
    fprintf(stderr, "A tool to construct homologous groups, please see README file for more details.\n  args: [F=?] [T=?] [S=?] [M=?] [C=?] tag-file\n\tT(15): number of refinement iterations\n\tS(10): number of splitting and merging iterations\n\tM(50): min length\n\tC(300): min chain\nWritten and maintained by Minmei Hou, PhD candidate, CSE Dept. Penn State Univ.\nPlease cite: ...\n");
    exit(1);
  }

  // self=1 by default
  while ( argv[1][1] == '=' ) {
    if ( argv[1][0]=='F' )
      self = atoi(argv[1]+2);
    else if ( argv[1][0] == 'T' )
      iter = atoi(argv[1]+2);
    else if ( argv[1][0] == 'S' )
      SPLITMERGE_ITERATION = atoi(argv[1]+2);
    else if ( argv[1][0] == 'M' )
      MIN_LENGTH = atoi(argv[1]+2);
    else if ( argv[1][0] == 'C' )
      MIN_CHAIN = atoi(argv[1]+2);
    else
      cerr << "wrong input\n";
    argv++;
    argc--;
  }

  initialize_chrs(argv[1]);
  alltags = readTagFile(argv[1]);
  
  sprintf(filename, "%s.dup", argv[1]);
  ofdup.open(filename);
  sprintf(filename, "%s.rep", argv[1]);
  ofrep.open(filename);
	
	tags = splitDupRepeat(alltags, ofdup, ofrep);
	ofdup.close();
	ofrep.close();
  
  (*tags)[0]->b1; // for using gdb purpose
  nodes[0]->start;
  getNodesFromTags(tags, nodes);
 
  // assign tags to appropriate node by x1, y1 values
  groupTags1(nodes, tags); // tags shall be empty now
  // get rid of small overlapping
  trimTags(nodes);
  checkTags(nodes);

  // cut nodes to make each node has one copy
  if ( self == 1 )
    addSelfAlignment(nodes);
  splitNodes(nodes);
  if ( self == 1 ) {
    rmSelfAlignment(nodes);
    maintenance_nodes(nodes);
  }
  cerr << endl;
  coverage = coverageLength(nodes);
  cerr << " node size after split: " << nodes.size() << ' ' << coverage << endl;
  // union neighboring nodes
  unionNodes(nodes, 50);
  checkTags(nodes);

  // get rid of small tags/chains
  remove_small_chains(nodes);
  maintenance_nodes(nodes);

  // compose edges
  construct_graph(nodes);
  checkTags(nodes);

  iter_count=0;
  int adjust=1;
  // tags shall be empty now
  while ( adjust>0 && iter_count<iter) {
    coverage = coverageLength(nodes);
    cerr << iter_count << " node size after union: " << nodes.size() << ' ' << coverage << endl;
    adjust = adjustLargeNodes(nodes);
    coverage = coverageLength(nodes);
    cerr << iter_count << " node size after split: " << nodes.size() << ' ' << coverage << endl;
    //remove_small_chains(nodes);
    maintenance_nodes(nodes);
    newsize = nodes.size();
    if ( iter_count >0 && newsize == oldsize )
      break;
    oldsize = newsize;
    //if ( iter_count < iter-1 )
      unionNodes(nodes, 50);

    checkTags(nodes);
    make_good_graph(nodes);
    iter_count++;
  }

  // split nodes if necessary to avoid fusing wrong nodes

  ccs = get_ccs(nodes);
  coverage = coverageLength(nodes);
  cerr << iter_count << " node size after union: " << nodes.size() << ' ' << coverage << " cc size: " << ccs->size() << endl;
  //finalizeNodes(nodes, ccs);
  delete_ccs(ccs);
  //remove_small_chains(nodes);
  maintenance_nodes(nodes);
  

  for (int i=0; i<5; i++) { // nodes number shall change
    oldsize = nodes.size();
    mergeSplitIteration(nodes, mergeGraph);
    newsize = nodes.size();
    if ( newsize == oldsize )
      break;
  }
  for (int i=0; i<SPLITMERGE_ITERATION; i++) { // ccs number shall change
    newsize = mergeSplitIteration(nodes, splitGraph);
    if ( i>0 && newsize == oldsize )
      break;
    mergeSplitIteration(nodes, mergeGraph);
    mergeSplitIteration(nodes, mergeGraph);
    oldsize = newsize;
  }

  ccs = get_ccs(nodes);
  sortCCS(ccs);
  collapse(*ccs, nodes, 3000);
  delete_ccs(ccs);
  maintenance_nodes(nodes);

  make_good_graph(nodes);
  program_display(nodes, argv[1]);

  clear_nodes(nodes);

  free_chrs();
  return 0;
}
